#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorBinOp, TensorConstant


class TensorLshift(operands.Lshift, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLshift, self).__init__(_casting=casting, _err=err,
                                           _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return x1.issparse() or x2.issparse()

    @classmethod
    def constant_cls(cls):
        return TensorLshiftConstant


class TensorLshiftConstant(operands.RshiftConstant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLshiftConstant, self).__init__(_casting=casting, _err=err,
                                                   _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if (hasattr(x1, 'issparse') and x1.issparse()) or \
                (hasattr(x2, 'issparse') and x2.issparse()):
            return True
        return False


@infer_dtype(np.left_shift)
def lshift(x1, x2, out=None, where=None, **kwargs):
    """
    Shift the bits of an integer to the left.

    Bits are shifted to the left by appending `x2` 0s at the right of `x1`.
    Since the internal representation of numbers is in binary format, this
    operation is equivalent to multiplying `x1` by ``2**x2``.

    Parameters
    ----------
    x1 : array_like of integer type
        Input values.
    x2 : array_like of integer type
        Number of zeros to append to `x1`. Has to be non-negative.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : tensor of integer type
        Return `x1` with bits shifted `x2` times to the left.

    See Also
    --------
    right_shift : Shift the bits of an integer to the right.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.left_shift(5, 2).execute()
    20

    >>> mt.left_shift(5, [1,2,3]).execute()
    array([10, 20, 40])
    """
    op = TensorLshift(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.left_shift, reverse=True)
def rlshift(x1, x2, **kwargs):
    op = TensorLshift(**kwargs)
    return op.rcall(x1, x2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Container, Iterable, Sequence
from functools import wraps
from math import factorial

import numpy as np

from ...compat import getargspec, numpy_compat
from ...operands import Mean
from .array import get_array_module, as_same_device, device, cp


def keepdims_wrapper(a_callable):
    @wraps(a_callable)
    def keepdims_wrapped_callable(x, axis=None, keepdims=None, *args, **kwargs):
        xp = get_array_module(x)
        if xp == np:
            func = a_callable
        else:
            func = getattr(xp, a_callable.__name__)

        r = func(x, axis=axis, *args, **kwargs)

        if not keepdims:
            return xp.asarray(r)

        axes = axis

        if axes is None:
            axes = range(x.ndim)

        if not isinstance(axes, (Container, Iterable, Sequence)):
            axes = [axes]

        if r.ndim != x.ndim:
            r_slice = tuple()
            for each_axis in range(x.ndim):
                if each_axis in axes:
                    r_slice += (None,)
                else:
                    r_slice += (slice(None),)

            r = r[r_slice]

        return r

    return keepdims_wrapped_callable

sum = keepdims_wrapper(np.sum)
prod = keepdims_wrapper(np.prod)
min = keepdims_wrapper(np.min)
max = keepdims_wrapper(np.max)
argmin = keepdims_wrapper(np.argmin)
nanargmin = keepdims_wrapper(np.nanargmin)
argmax = keepdims_wrapper(np.argmax)
nanargmax = keepdims_wrapper(np.nanargmax)
any = keepdims_wrapper(np.any)
all = keepdims_wrapper(np.all)
nansum = keepdims_wrapper(np.nansum)

try:
    from numpy import nanprod, nancumprod, nancumsum
except ImportError:  # pragma: no cover
    nanprod = numpy_compat.nanprod
    nancumprod = numpy_compat.nancumprod
    nancumsum = numpy_compat.nancumsum

nanprod = keepdims_wrapper(nanprod)
nancumprod = keepdims_wrapper(nancumprod)
nancumsum = keepdims_wrapper(nancumsum)

nanmin = keepdims_wrapper(np.nanmin)
nanmax = keepdims_wrapper(np.nanmax)
mean = keepdims_wrapper(np.mean)


def _handle_reduction(op):
    def _handle(ctx, chunk):
        (input,), device_id, _ = as_same_device(
            [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)
        axis = _get_axis(chunk.op.axis)
        with device(device_id):
            if "dtype" in getargspec(op).args:
                ctx[chunk.key] = op(input, axis=axis,
                                    dtype=chunk.op.dtype,
                                    keepdims=bool(chunk.op.keepdims))
            else:
                ctx[chunk.key] = op(input, axis=axis,
                                    keepdims=bool(chunk.op.keepdims))
    return _handle


def _mean_chunk(ctx, chunk):
    from ..expressions.reduction import TensorNanMeanChunk

    (in_chunk,), device_id, _ = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    axis = _get_axis(chunk.op.axis)
    _count_func = _nannumel if (type(chunk.op) is TensorNanMeanChunk) else _numel
    _sum_func = nansum if (type(chunk.op) is TensorNanMeanChunk) else sum

    with device(device_id):
        chunk_count = _count_func(in_chunk, axis=axis, dtype=np.int64,
                                  keepdims=bool(chunk.op.keepdims))
        chunk_sum = _sum_func(in_chunk, axis=axis, dtype=chunk.op.dtype,
                              keepdims=bool(chunk.op.keepdims))
        ctx[chunk.key] = (chunk_sum, chunk_count)


def _mean_combine(ctx, chunk):
    axis = _get_axis(chunk.op.axis)
    (_data, _count), device_id, _ = as_same_device(
        ctx[chunk.inputs[0].key], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        chunk_count = sum(_count, axis=axis, dtype=np.int64,
                          keepdims=bool(chunk.op.keepdims))
        chunk_sum = sum(_data, axis=axis, dtype=chunk.op.dtype,
                        keepdims=bool(chunk.op.keepdims))
        ctx[chunk.key] = (chunk_sum, chunk_count)


def _mean(ctx, chunk):
    axis = _get_axis(chunk.op.axis)

    a = ctx[chunk.inputs[0].key]
    if not isinstance(a, (list, tuple)):
        (inp,), device_id, xp = as_same_device(
            [a], device=chunk.op.device, ret_extra=True)

        with device(device_id):
            func = xp.mean if isinstance(chunk.op, Mean) else xp.nanmean
            ctx[chunk.key] = func(inp, axis=axis, dtype=chunk.op.dtype,
                                  keepdims=bool(chunk.op.keepdims))
    else:
        (_data, _count), device_id, xp = as_same_device(
            a, device=chunk.op.device, ret_extra=True)

        with device(device_id):
            chunk_count = sum(_count, axis=axis, dtype=chunk.op.dtype,
                              keepdims=bool(chunk.op.keepdims))
            chunk_sum = sum(_data, axis=axis, dtype=chunk.op.dtype,
                            keepdims=bool(chunk.op.keepdims))
            ctx[chunk.key] = xp.true_divide(chunk_sum, chunk_count,
                                            dtype=chunk.op.dtype)


def _handle_arg_chunk(op, arg_op):
    def _handle(ctx, chunk):
        arg_axis = _get_arg_axis(chunk.op.axis, chunk.inputs[0].ndim)
        (in_chunk,), device_id, xp = as_same_device(
            [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)
        offset = chunk.op.offset

        with device(device_id):
            vals = op(in_chunk, axis=arg_axis, keepdims=True)
            arg = arg_op(in_chunk, axis=arg_axis, keepdims=True)

            if arg_axis is None:
                if xp == cp:
                    # we need to copy to do cpu computation, then copy back to gpu
                    # cuz unravel_index and ravel_multi_index are not implemented in cupy
                    in_chunk = in_chunk.get()

                total_shape = chunk.op.total_shape
                ind = np.unravel_index(arg.ravel()[0], in_chunk.shape)
                total_ind = tuple(o + i for (o, i) in zip(offset, ind))
                res = np.ravel_multi_index(total_ind, total_shape)

                if xp == cp:
                    # copy back
                    with xp.cuda.Device(in_chunk.device.id):
                        arg[:] = xp.asarray(res)
                else:
                    arg[:] = res
            else:
                arg += offset
            ctx[chunk.key] = (vals, arg)
    return _handle


def _handle_arg_combine(arg_op):
    def _handle(ctx, chunk):
        axis = _get_arg_axis(chunk.op.axis, chunk.inputs[0].ndim)
        (vals, arg), device_id, xp = as_same_device(
            ctx[chunk.inputs[0].key], device=chunk.op.device, ret_extra=True)
        keepdims = bool(chunk.op.keepdims)

        with device(device_id):
            if axis is None:
                local_args = arg_op(vals, axis=axis, keepdims=keepdims)
                vals = vals.ravel()[local_args]
                arg = arg.ravel()[local_args]
            else:
                local_args = arg_op(vals, axis=axis)
                inds = np.ogrid[tuple(map(slice, local_args.shape))]
                if xp != np:
                    inds = [xp.asarray(it) for it in inds]
                inds.insert(axis, local_args)
                inds_tuple = tuple(inds)
                vals = vals[inds_tuple]
                arg = arg[inds_tuple]
                if keepdims:
                    vals = xp.expand_dims(vals, axis)
                    arg = xp.expand_dims(arg, axis)
            ctx[chunk.key] = (vals, arg)
    return _handle


def _handle_arg(arg_op):
    def _handle(ctx, chunk):
        axis = _get_arg_axis(chunk.op.axis, chunk.inputs[0].ndim)
        (vals, arg), device_id, xp = as_same_device(
            ctx[chunk.inputs[0].key], device=chunk.op.device, ret_extra=True)

        with device(device_id):
            if xp.any(xp.isnan(vals)) and arg_op in [_nanargmin, _nanargmax]:
                raise ValueError("All NaN slice encountered")
            keepdims = bool(chunk.op.keepdims)
            if axis is None:
                local_args = arg_op(vals, axis=axis, keepdims=keepdims)
                arg = arg.ravel()[local_args]
            else:
                local_args = arg_op(vals, axis=axis)
                inds = np.ogrid[tuple(map(slice, local_args.shape))]
                if xp != np:
                    inds = [xp.asarray(it) for it in inds]
                inds.insert(axis, local_args)
                arg = arg[tuple(inds)]
                if keepdims:
                    arg = xp.expand_dims(arg, axis)
            ctx[chunk.key] = arg
    return _handle


def _handle_cum(cum_op):
    def _handle(ctx, chunk):
        (x,), device_id, xp = as_same_device(
            [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

        if xp != np:
            func = getattr(xp, cum_op.__name__)
        else:
            func = cum_op

        with device(device_id):
            ctx[chunk.key] = func(x, axis=chunk.op.axis,
                                  dtype=chunk.op.dtype)
    return _handle


def _get_axis(axis):
    return tuple(axis) if axis is not None else axis


def _get_arg_axis(axis, ndim):
    return None if len(axis) == ndim or ndim == 1 else axis[0]


def _numel(x, **kwargs):
    xp = get_array_module(x)
    return sum(xp.ones_like(x), **kwargs)


def _nannumel(x, **kwargs):
    xp = get_array_module(x)
    return sum(~xp.isnan(x), **kwargs)


def _nanargmin(x, **kwargs):
    try:
        return nanargmin(x, **kwargs)
    except ValueError:
        xp = get_array_module(x)
        return nanargmin(xp.where(xp.isnan(x), np.inf, x), **kwargs)


def _nanargmax(x, axis, **kwargs):
    try:
        return nanargmax(x, axis, **kwargs)
    except ValueError:
        xp = get_array_module(x)
        return nanargmax(xp.where(xp.isnan(x), -np.inf, x), **kwargs)


def _count_nonzero(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], chunk.op.device, ret_extra=True)

    axis = _get_arg_axis(chunk.op.axis, chunk.inputs[0].ndim)
    keepdims = chunk.op.keepdims
    with device(device_id):
        nz = xp.count_nonzero(x, axis=axis)
        if keepdims:
            slcs = [slice(None)] * chunk.inputs[0].ndim
            for ax in chunk.op.axis:
                slcs[ax] = np.newaxis
            nz = xp.asarray(nz)[tuple(slcs)]

        ctx[chunk.key] = nz


def _moment_chunk(ctx, chunk):
    from ..expressions.reduction import TensorNanMomentChunk

    (in_chunk,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    axis = _get_axis(chunk.op.axis)
    moment = chunk.op.moment
    dtype = chunk.op.dtype
    _count_func = _nannumel if type(chunk.op) is TensorNanMomentChunk else _numel
    _sum_func = nansum if type(chunk.op) is TensorNanMomentChunk else sum
    empty = get_array_module(in_chunk, nosparse=True).empty

    with device(device_id):
        chunk_count = _count_func(in_chunk, axis=axis, dtype=np.int64,
                                  keepdims=bool(chunk.op.keepdims))
        chunk_sum = _sum_func(in_chunk, axis=axis, dtype=dtype,
                              keepdims=bool(chunk.op.keepdims))
        avg = xp.true_divide(chunk_sum, chunk_count)
        var_square = empty(chunk_count.shape + (moment - 1,), dtype=dtype)
        for i in range(2, moment + 1):
            var_square[..., i - 2] = _sum_func((in_chunk - avg) ** i, axis=axis, dtype=dtype,
                                               keepdims=bool(chunk.op.keepdims))
        ctx[chunk.key] = (chunk_sum, chunk_count, var_square)


def _reduce_var_square(var_square, avg_diff, count, chunk, axis, sum_func):
    moment = chunk.op.moment
    dtype = chunk.op.dtype
    kw = dict(axis=axis, dtype=dtype, keepdims=bool(chunk.op.keepdims))

    reduced_var_square = var_square[..., moment - 2].sum(**kw) + \
                         sum_func(count * avg_diff ** moment, **kw)
    for i in range(1, moment - 1):
        coeff = factorial(moment) / float(factorial(i) * factorial(moment - i))
        reduced_var_square += coeff * sum_func(var_square[..., moment - i - 2] * avg_diff ** moment, **kw)
    return reduced_var_square


def _moment_combine(ctx, chunk):
    from ..expressions.reduction import TensorNanMomentCombine

    axis = _get_axis(chunk.op.axis)
    moment = chunk.op.moment
    dtype = chunk.op.dtype
    _sum_func = nansum if type(chunk.op) is TensorNanMomentCombine else sum

    (_data, _count, _var_square), device_id, xp = as_same_device(
        ctx[chunk.inputs[0].key], device=chunk.op.device, ret_extra=True)
    empty = get_array_module(_data, nosparse=True).empty

    with device(device_id):
        chunk_count = _sum_func(_count, axis=axis, dtype=np.int64,
                                keepdims=bool(chunk.op.keepdims))
        chunk_sum = _sum_func(_data, axis=axis, dtype=dtype,
                              keepdims=bool(chunk.op.keepdims))
        avg = xp.true_divide(chunk_sum, chunk_count, dtype=dtype)
        avg_diff = xp.true_divide(_data, _count, dtype=dtype) - avg
        var_square = empty(chunk_count.shape + (moment - 1,), dtype=dtype)

        for m in range(2, moment + 1):
            var_square[..., m - 2] = \
                _reduce_var_square(_var_square, avg_diff, _count, chunk, axis, _sum_func)

        ctx[chunk.key] = (chunk_sum, chunk_count, var_square)


def _moment(ctx, chunk):
    from ..expressions.reduction import TensorNanMoment

    axis = _get_axis(chunk.op.axis)
    dtype = chunk.op.dtype
    _sum_func = nansum if type(chunk.op) is TensorNanMoment else sum

    (_data, _count, _var_square), device_id, xp = as_same_device(
        ctx[chunk.inputs[0].key], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        chunk_count = _sum_func(_count, axis=axis, dtype=np.int64,
                                keepdims=True)
        chunk_sum = _sum_func(_data, axis=axis, dtype=dtype, keepdims=True)
        avg = xp.true_divide(chunk_sum, chunk_count, dtype=dtype)
        avg_diff = xp.true_divide(_data, _count, dtype=dtype) - avg
        var_square = _reduce_var_square(_var_square, avg_diff, _count, chunk, axis, _sum_func)

        ctx[chunk.key] = xp.true_divide(
            var_square,
            sum(chunk_count, axis=axis, dtype=dtype, keepdims=bool(chunk.op.keepdims)) - chunk.op.ddof,
            dtype=dtype)


def _var(ctx, chunk):
    from ..expressions.reduction import TensorNanVar

    axis = _get_axis(chunk.op.axis)
    (in_chunk,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        func = xp.nanvar if type(chunk.op) is TensorNanVar else xp.var
        ctx[chunk.key] = func(in_chunk, axis=axis, dtype=chunk.op.dtype, ddof=chunk.op.ddof,
                              keepdims=bool(chunk.op.keepdims))


def register_reduction_handler():
    from .core import register
    from ...operands import Sum, Prod, Min, Max, All, Any, Mean, MeanChunk, MeanCombine, \
        Argmax, ArgmaxChunk, ArgmaxCombine, Argmin, ArgminChunk, ArgminCombine, NanSum,\
        NanProd, NanMax, NanMin, NanMeanChunk, NanMean, NanArgmax, NanArgmaxChunk, NanArgmin,\
        NanArgminChunk, NanArgmaxCombine, NanArgminCombine, Moment, NanMomentChunk, MomentChunk, \
        MomentCombine, NanMomentCombine, NanMoment, Var, NanVar, CountNonzero, \
        Cumsum, Cumprod, NanCumsum, NanCumprod

    register(Sum, _handle_reduction(sum))
    register(NanSum, _handle_reduction(nansum))
    register(Prod, _handle_reduction(prod))
    register(NanProd, _handle_reduction(nanprod))
    register(Max, _handle_reduction(max))
    register(NanMax, _handle_reduction(nanmax))
    register(Min, _handle_reduction(min))
    register(NanMin, _handle_reduction(nanmin))
    register(All, _handle_reduction(all))
    register(Any, _handle_reduction(any))

    register(Mean, _mean)
    register(NanMean, _mean)
    register(Var, _var)
    register(NanVar, _var)
    register(MeanChunk, _mean_chunk)
    register(NanMeanChunk, _mean_chunk)
    register(MeanCombine, _mean_combine)

    register(Moment, _moment)
    register(NanMoment, _moment)
    register(MomentChunk, _moment_chunk)
    register(NanMomentChunk, _moment_chunk)
    register(MomentCombine, _moment_combine)
    register(NanMomentCombine, _moment_combine)

    register(CountNonzero, _count_nonzero)

    register(Argmax, _handle_arg(argmax))
    register(NanArgmax, _handle_arg(_nanargmax))
    register(ArgmaxChunk, _handle_arg_chunk(max, argmax))
    register(NanArgmaxChunk, _handle_arg_chunk(nanmax, _nanargmax))
    register(ArgmaxCombine, _handle_arg_combine(argmax))
    register(NanArgmaxCombine, _handle_arg_combine(_nanargmax))
    register(Argmin, _handle_arg(argmin))
    register(NanArgmin, _handle_arg(_nanargmin))
    register(ArgminChunk, _handle_arg_chunk(min, argmin))
    register(NanArgminChunk, _handle_arg_chunk(nanmin, _nanargmin))
    register(ArgminCombine, _handle_arg_combine(argmin))
    register(NanArgminCombine, _handle_arg_combine(_nanargmin))

    register(Cumsum, _handle_cum(np.cumsum))
    register(Cumprod, _handle_cum(np.cumprod))
    register(NanCumsum, _handle_cum(nancumsum))
    register(NanCumprod, _handle_cum(nancumprod))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gevent.monkey
gevent.monkey.patch_all(thread=False)

import logging  # noqa: E402
import random   # noqa: E402
import time     # noqa: E402

from ..base_app import BaseApplication   # noqa: E402
from ..compat import six                 # noqa: E402
from ..errors import StartArgumentError  # noqa: E402

logger = logging.getLogger(__name__)


class WebApplication(BaseApplication):
    def __init__(self):
        super(WebApplication, self).__init__()
        self.mars_web = None
        self.require_pool = False

    def config_args(self, parser):
        parser.add_argument('--ui-port', help='port of Mars UI')

    def validate_arguments(self):
        if not self.args.schedulers and not self.args.kv_store:
            raise StartArgumentError('Either schedulers or url of kv store is required.')

    def main_loop(self):
        try:
            self.start_service()
            while True:
                time.sleep(0.1)
        finally:
            self.stop_service()

    def start_service(self):
        from .server import MarsWeb
        if MarsWeb is None:
            self.mars_web = None
            logger.warning('Mars UI cannot be loaded. Please check if necessary components are installed.')
        else:
            ui_port = int(self.args.ui_port) if self.args.ui_port else None
            scheduler_ip = self.args.schedulers or None
            if isinstance(scheduler_ip, six.string_types):
                schedulers = scheduler_ip.split(',')
                scheduler_ip = schedulers[random.randint(0, len(schedulers) - 1)]
            self.mars_web = MarsWeb(port=ui_port, scheduler_ip=scheduler_ip)
            self.mars_web.start()

    def stop_service(self):
        if self.mars_web:
            self.mars_web.stop()


main = WebApplication()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorUnaryOp


class TensorExpm1(operands.Expm1, TensorUnaryOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorExpm1, self).__init__(_casting=casting, _err=err,
                                          _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x):
        return False


@infer_dtype(np.expm1)
def expm1(x, out=None, where=None, **kwargs):
    """
    Calculate ``exp(x) - 1`` for all elements in the tensor.

    Parameters
    ----------
    x : array_like
       Input values.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : Tensor
        Element-wise exponential minus one: ``out = exp(x) - 1``.

    See Also
    --------
    log1p : ``log(1 + x)``, the inverse of expm1.


    Notes
    -----
    This function provides greater precision than ``exp(x) - 1``
    for small values of ``x``.

    Examples
    --------
    The true value of ``exp(1e-10) - 1`` is ``1.00000000005e-10`` to
    about 32 significant digits. This example shows the superiority of
    expm1 in this case.

    >>> import mars.tensor as mt

    >>> mt.expm1(1e-10).execute()
    1.00000000005e-10
    >>> (mt.exp(1e-10) - 1).execute()
    1.000000082740371e-10
    """
    op = TensorExpm1(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ...compat import izip
from ...compat.numpy_compat import broadcast_to
from .array import as_same_device, device


def _copyto(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        dst = inputs[1].copy()
        src = inputs[0]
        where = inputs[2] if len(inputs) > 2 else None

        xp.copyto(dst, src, casting=chunk.op.casting, where=where)
        ctx[chunk.key] = dst


def _astype(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        if chunk.op.sparse:
            ctx[chunk.key] = x.astype(chunk.op.dtype)
        else:
            ctx[chunk.key] = x.astype(chunk.op.dtype, casting=chunk.op.casting)


def _transpose(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    axes = chunk.op.axes
    with device(device_id):
        ctx[chunk.key] = xp.transpose(x, axes or None)


def _swapaxes(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    axis1, axis2 = chunk.op.axis1, chunk.op.axis2
    with device(device_id):
        ctx[chunk.key] = xp.swapaxes(x, axis1, axis2)


def _broadcast_to(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    if xp == np:
        func = broadcast_to
    else:
        func = xp.broadcast_to

    with device(device_id):
        ctx[chunk.key] = func(x, chunk.op.shape)


def _where(ctx, chunk):
    (cond, x, y), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.where(cond, x, y)


def _reshape(ctx, chunk):
    (x,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = x.reshape(chunk.op.newshape)


def _split(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    a = inputs[0]
    ind = inputs[1] if len(inputs) > 1 else chunk.op.indices_or_sections
    with device(device_id):
        ret = xp.array_split(a, ind)
        assert len(chunk.op.outputs) == len(ret)
        for o, r in izip(chunk.op.outputs, ret):
            ctx[o.key] = r


def _squeeze(ctx, chunk):
    (a,), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.squeeze(a, axis=chunk.op.axis)


def _digitize(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    x = inputs[0]
    if len(inputs) > 1:
        bins = inputs[1]
    else:
        bins = chunk.op.bins

    with device(device_id):
        ctx[chunk.key] = xp.digitize(x, bins=bins, right=chunk.op.right)


def _repeat(ctx, chunk):
    inputs, device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    a = inputs[0]
    if len(inputs) > 1:
        repeats = inputs[1]
    else:
        repeats = chunk.op.repeats

    with device(device_id):
        ctx[chunk.key] = xp.repeat(a, repeats=repeats, axis=chunk.op.axis)


def _isin(ctx, chunk):
    (element, test_elements), device_id, xp = as_same_device(
        [ctx[c.key] for c in chunk.inputs], device=chunk.op.device, ret_extra=True)

    with device(device_id):
        ctx[chunk.key] = xp.isin(element, test_elements,
                                 assume_unique=chunk.op.assume_unique,
                                 invert=chunk.op.invert)


def register_basic_handler():
    from ... import operands
    from .core import register

    register(operands.CopyTo, _copyto)
    register(operands.Astype, _astype)
    register(operands.Transpose, _transpose)
    register(operands.SwapAxes, _swapaxes)
    register(operands.BroadcastTo, _broadcast_to)
    register(operands.Where, _where)
    register(operands.Reshape, _reshape)
    register(operands.Split, _split)
    register(operands.Squeeze, _squeeze)
    register(operands.Digitize, _digitize)
    register(operands.Repeat, _repeat)
    register(operands.IsIn, _isin)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
from numbers import Number

import numpy as np

from .... import operands
from ...core import TENSOR_TYPE, CHUNK_TYPE, Tensor
from ..utils import broadcast_shape
from ..datasource import tensor as astensor
from .core import TensorElementWise


class TensorClip(operands.Clip, TensorElementWise):
    @contextlib.contextmanager
    def _handle_params(self, inputs):
        inps = inputs[:1]
        inputs_iter = iter(inputs[1:])

        if getattr(self, '_a', None) is None:
            # create clip op from beginning

            # a_min
            has_a_min = False
            a_min = next(inputs_iter)
            if a_min is not None and isinstance(a_min, TENSOR_TYPE + CHUNK_TYPE):
                has_a_min = True
                inps.append(a_min)
            else:
                setattr(self, '_a_min', a_min)

            # a_max
            has_a_max = False
            a_max = next(inputs_iter)
            if a_max is not None and isinstance(a_max, TENSOR_TYPE + CHUNK_TYPE):
                has_a_max = True
                inps.append(a_max)
            else:
                setattr(self, '_a_max', a_max)

            # out
            has_out = False
            out = next(inputs_iter)
            if out is not None:
                has_out = True
                inps.append(out)
        else:
            # create clip op from existence

            # a_min
            raw_a_min = getattr(self, '_a_min', None)
            has_a_min = raw_a_min is not None and isinstance(raw_a_min, TENSOR_TYPE + CHUNK_TYPE)
            if has_a_min:
                a_min = next(inputs_iter)
                inps.append(a_min)

            # a_max
            raw_a_max = getattr(self, '_a_max', None)
            has_a_max = raw_a_max is not None and isinstance(raw_a_max, TENSOR_TYPE + CHUNK_TYPE)
            if has_a_max:
                a_max = next(inputs_iter)
                inps.append(a_max)

            # out
            has_out = getattr(self, '_out', None) is not None
            if has_out:
                out = next(inputs_iter)
                inps.append(out)

        yield inps

        inputs = getattr(self, '_inputs')
        inputs_iter = iter(inputs)
        setattr(self, '_a', next(inputs_iter))
        if has_a_min:
            setattr(self, '_a_min', next(inputs_iter))
        if has_a_max:
            setattr(self, '_a_max', next(inputs_iter))
        if has_out:
            setattr(self, '_out', next(inputs_iter))

    def new_tensors(self, inputs, shape,**kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorClip, self).new_tensors(inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorClip, self).new_chunks(inputs, shape, **kw)

    def __call__(self, a, a_min, a_max, out=None):
        a = astensor(a)
        tensors = [a]
        sparse = a.issparse()

        if isinstance(a_min, Number):
            if a_min > 0:
                sparse = False
            a_min_dtype = np.array(a_min).dtype
        else:
            a_min = astensor(a_min)
            tensors.append(a_min)
            if not a_min.issparse():
                sparse = False
            a_min_dtype = a_min.dtype

        if isinstance(a_max, Number):
            if a_max < 0:
                sparse = False
            a_max_dtype = np.array(a_max).dtype
        else:
            a_max = astensor(a_max)
            tensors.append(a_max)
            if not a_max.issparse():
                sparse = False
            a_max_dtype = a_max.dtype

        dtype = np.result_type(a.dtype, a_min_dtype, a_max_dtype)
        # check broadcast
        shape = broadcast_shape(*[t.shape for t in tensors])

        setattr(self, '_sparse', sparse)
        t = self.new_tensor([a, a_min, a_max, out], shape)

        if out is None:
            setattr(self, '_dtype', dtype)
            return t

        if not isinstance(out, Tensor):
            raise TypeError('out should be Tensor object, got {0} instead'.format(type(out)))
        out_shape, out_dtype = out.shape, out.dtype

        # if `out` is specified, use out's dtype and shape
        if t.shape != out_shape:
            t = self.new_tensor([a, a_min, a_max, out], out_shape)
        setattr(self, '_dtype', out_dtype)

        out.data = t.data
        return out


def clip(a, a_min, a_max, out=None):
    """
    Clip (limit) the values in a tensor.

    Given an interval, values outside the interval are clipped to
    the interval edges.  For example, if an interval of ``[0, 1]``
    is specified, values smaller than 0 become 0, and values larger
    than 1 become 1.

    Parameters
    ----------
    a : array_like
        Tensor containing elements to clip.
    a_min : scalar or array_like or `None`
        Minimum value. If `None`, clipping is not performed on lower
        interval edge. Not more than one of `a_min` and `a_max` may be
        `None`.
    a_max : scalar or array_like or `None`
        Maximum value. If `None`, clipping is not performed on upper
        interval edge. Not more than one of `a_min` and `a_max` may be
        `None`. If `a_min` or `a_max` are array_like, then the three
        arrays will be broadcasted to match their shapes.
    out : Tensor, optional
        The results will be placed in this tensor. It may be the input
        array for in-place clipping.  `out` must be of the right shape
        to hold the output.  Its type is preserved.

    Returns
    -------
    clipped_array : Tensor
        An tensor with the elements of `a`, but where values
        < `a_min` are replaced with `a_min`, and those > `a_max`
        with `a_max`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.arange(10)
    >>> mt.clip(a, 1, 8).execute()
    array([1, 1, 2, 3, 4, 5, 6, 7, 8, 8])
    >>> a.execute()
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> mt.clip(a, 3, 6, out=a).execute()
    array([3, 3, 3, 3, 4, 5, 6, 6, 6, 6])
    >>> a = mt.arange(10)
    >>> a.execute()
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> mt.clip(a, [3, 4, 1, 1, 1, 4, 4, 4, 4, 4], 8).execute()
    array([3, 4, 2, 3, 4, 5, 6, 7, 8, 8])

    """
    op = TensorClip()
    return op(a, a_min, a_max, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number

import numpy as np

from ..datasource import tensor as astensor
from .. import arithmetic as arith


UFUNC_TO_TENSOR_FUNC = {
    # binary
    np.add: arith.add,
    np.subtract: arith.subtract,
    np.multiply: arith.multiply,
    np.divide: arith.divide,
    np.logaddexp: arith.logaddexp,
    np.logaddexp2: arith.logaddexp2,
    np.true_divide: arith.truediv,
    np.floor_divide: arith.floordiv,
    # unary
    np.negative: arith.negative,
    np.power: arith.power,
    np.float_power: arith.float_power,
    np.remainder: arith.remainder,
    np.mod: arith.mod,
    np.fmod: arith.fmod,
    np.conj: arith.conj,
    np.conjugate: arith.conjugate,
    np.exp: arith.exp,
    np.exp2: arith.exp2,
    np.log: arith.log,
    np.log2: arith.log2,
    np.log10: arith.log10,
    np.log1p: arith.log1p,
    np.expm1: arith.expm1,
    np.sqrt: arith.sqrt,
    np.square: arith.square,
    np.cbrt: arith.cbrt,
    np.reciprocal: arith.reciprocal,
    # trigonometric functions
    np.sin: arith.sin,
    np.cos: arith.cos,
    np.tan: arith.tan,
    np.arcsin: arith.arcsin,
    np.arccos: arith.arccos,
    np.arctan: arith.arctan,
    np.arctan2: arith.arctan2,
    np.hypot: arith.hypot,
    np.sinh: arith.sinh,
    np.cosh: arith.cosh,
    np.tanh: arith.tanh,
    np.arcsinh: arith.arcsinh,
    np.arccosh: arith.arccosh,
    np.arctanh: arith.arctanh,
    np.deg2rad: arith.deg2rad,
    np.rad2deg: arith.rad2deg,
    # comparison functions
    np.greater: arith.greater,
    np.greater_equal: arith.greater_equal,
    np.less: arith.less,
    np.less_equal: arith.less_equal,
    np.not_equal: arith.not_equal,
    np.equal: arith.equal,
    np.logical_and: arith.logical_and,
    np.logical_or: arith.logical_or,
    np.logical_xor: arith.logical_xor,
    np.logical_not: arith.logical_not,
    np.maximum: arith.maximum,
    np.minimum: arith.minimum,
    np.fmax: arith.fmax,
    np.fmin: arith.fmin,
    # floating functions
    np.isfinite: arith.isfinite,
    np.isinf: arith.isinf,
    np.isnan: arith.isnan,
    np.signbit: arith.signbit,
    np.copysign: arith.copysign,
    np.nextafter: arith.nextafter,
    np.spacing: arith.spacing,
    np.modf: arith.modf,
    np.ldexp: arith.ldexp,
    np.frexp: arith.frexp,
    np.floor: arith.floor,
    np.ceil: arith.ceil,
    np.trunc: arith.trunc,
    # more math functions
    np.degrees: arith.degrees,
    np.radians: arith.radians,
    np.rint: arith.rint,
    np.fabs: arith.fabs,
    np.sign: arith.sign,
    np.absolute: arith.absolute,
}


def _check_arg(arg):
    if isinstance(arg, Number):
        return True

    try:
        astensor(arg)
        return True
    except ValueError:
        return False


def _array_ufunc(_, ufunc, method, *inputs, **kwargs):
    out = kwargs.get('out', tuple())
    for x in inputs + out:
        if not _check_arg(x):
            return NotImplemented

    if method == '__call__':
        if ufunc.signature is not None:
            return NotImplemented
        if ufunc not in UFUNC_TO_TENSOR_FUNC:
            return NotImplemented

        tensor_func = UFUNC_TO_TENSOR_FUNC[ufunc]
        return tensor_func(*inputs, **kwargs)

    return NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral
import operator
import itertools
import contextlib

import numpy as np

from ....operands import Index
from ....core import BaseWithKey, Entity
from ....compat import reduce, irange, izip
from ...core import TENSOR_TYPE
from ..utils import unify_chunks, split_index_into_chunks, is_asc_sorted, \
    slice_split, calc_sliced_size
from ..core import TensorOperandMixin
from .core import process_index, get_index_and_shape


class TensorIndex(Index, TensorOperandMixin):
    def __init__(self, dtype=None, sparse=False, **kw):
        super(TensorIndex, self).__init__(_dtype=dtype, _sparse=sparse, **kw)

    @contextlib.contextmanager
    def _handle_params(self, inputs, indexes):
        """
        Index operator is special, it has additional parameter `indexes` which may also be tensor type,
        normally, this indexes is provided when called by `tile` or `TensorIndex.__call__`, however, calls
        in `GraphActor.get_executable_operand_dag` only provide inputs, in such situation, we need get `indexes`
        from operand itself and replace tensor-liked indexes by new one in `inputs`.
        """
        if indexes is not None:
            indexes_inputs = [ind for ind in indexes if isinstance(ind, (BaseWithKey, Entity))]
            inputs = inputs + indexes_inputs
        yield inputs

        if indexes is not None:
            self._indexes = indexes

        inputs_iter = iter(self._inputs[1:])
        new_indexes = [next(inputs_iter) if isinstance(index, (BaseWithKey, Entity)) else index
                       for index in self._indexes]
        self._indexes = new_indexes

    def new_tensors(self, inputs, shape, **kw):
        indexes = kw.pop('indexes', None)
        with self._handle_params(inputs, indexes) as mix_inputs:
            return super(TensorIndex, self).new_tensors(mix_inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        indexes = kw.pop('indexes', None)
        with self._handle_params(inputs, indexes) as mix_inputs:
            return super(TensorIndex, self).new_chunks(mix_inputs, shape, **kw)

    def __call__(self, a, index, shape):
        return self.new_tensor([a], shape, indexes=index)

    @classmethod
    def tile(cls, op):
        from ..merge.concatenate import TensorConcatenate

        in_tensor = op.input
        tensor = op.outputs[0]

        indexes = list(op.indexes)
        axis = 0
        output_axis = 0
        output_chunk_shape = []
        to_concat_axis_index = []
        for i, index in enumerate(indexes):
            if isinstance(index, TENSOR_TYPE) and index.dtype == np.bool_:
                # bool indexing
                # do unify chunk here
                in_tensor, index = unify_chunks(
                    in_tensor, (index, tuple(axis+ii for ii in range(index.ndim))))

                output_chunk_shape.append(reduce(operator.mul, index.chunk_shape))
                indexes[i] = index.chunks
                axis += index.ndim
                output_axis += 1
            elif isinstance(index, np.ndarray):
                # fancy indexing
                if index.ndim > 1:
                    # currently we only support tensor from numpy.ndarray
                    raise NotImplementedError
                try:
                    np_index = np.sort(index)
                    splits = split_index_into_chunks(in_tensor.nsplits[axis], np_index)
                except IndexError as e:
                    idx, size = e.idx, e.size
                    raise IndexError('index {0} is out of bounds for axis {1} with size {2}'.format(
                        idx, axis, size
                    ))

                non_empty_idx_splits = [(j, s) for j, s in enumerate(splits) if s.size > 0]
                non_empty_splits, _ = tuple(zip(*non_empty_idx_splits))
                try:
                    indexes[i] = non_empty_idx_splits
                except:
                    raise
                if not is_asc_sorted(index):
                    pos_index = np.searchsorted(np_index, index)
                    to_concat_axis_index.append((output_axis, pos_index))

                axis += 1
                output_axis += 1
                output_chunk_shape.append(len(non_empty_splits))
            elif isinstance(index, (slice, Integral)):
                indexes[i] = sorted(slice_split(index, in_tensor.nsplits[axis]).items(),
                                    key=operator.itemgetter(0))
                if isinstance(index, slice) and index.step is not None and index.step < 0:
                    indexes[i] = list(reversed(indexes[i]))
                if isinstance(index, slice):
                    output_chunk_shape.append(len(indexes[i]))
                axis += 1
                if isinstance(index, slice):
                    output_axis += 1
            elif isinstance(index, TENSOR_TYPE):
                raise NotImplementedError('Mars currently does not support fancy index from tensor')
            else:
                assert index is None
                output_chunk_shape.append(1)
                output_axis += 1

        out_chunks = []
        for output_idx in itertools.product(*(irange(s) for s in output_chunk_shape)):
            chunk_idx = []
            chunk_index = []  # chunk[index]
            chunk_shape = []
            axis = 0
            output_axis = 0
            for raw_index, index in izip(op.indexes, indexes):
                if isinstance(raw_index, TENSOR_TYPE) and raw_index.dtype == np.bool_:
                    indexed = index[output_idx[output_axis]]
                    chunk_index.append(indexed)
                    chunk_idx.extend(indexed.index)
                    chunk_shape.append(np.nan)
                    axis += raw_index.ndim
                    output_axis += 1
                elif isinstance(raw_index, np.ndarray):
                    input_index, indexed = index[output_idx[output_axis]]
                    chunk_index.append(indexed)
                    chunk_idx.append(input_index)
                    chunk_shape.append(len(indexed))
                    axis += 1
                    output_axis += 1
                elif isinstance(raw_index, slice):
                    sliceobj = index[output_idx[output_axis]][1]
                    chunk_index.append(sliceobj)
                    ix = index[output_idx[output_axis]][0]
                    chunk_idx.append(ix)
                    chunk_shape.append(calc_sliced_size(in_tensor.nsplits[axis][ix], sliceobj))
                    axis += 1
                    output_axis += 1
                elif isinstance(raw_index, Integral):
                    input_index, sliceobj = index[0]
                    chunk_index.append(sliceobj)
                    chunk_idx.append(input_index)
                    axis += 1
                else:
                    chunk_index.append(index)
                    chunk_shape.append(1)
                    output_axis += 1

            chunk_input = in_tensor.cix[tuple(chunk_idx)]
            chunk_op = op.copy().reset_key()
            chunk = chunk_op.new_chunk([chunk_input], tuple(chunk_shape), indexes=chunk_index, index=output_idx)
            out_chunks.append(chunk)

        nsplits = [tuple(c.shape[i] for c in out_chunks
                         if all(idx == 0 for j, idx in enumerate(c.index) if j != i))
                   for i in range(len(out_chunks[0].shape))]
        new_op = op.copy().reset_key()
        tensor = new_op.new_tensor([op.input], tensor.shape, indexes=op.indexes, chunks=out_chunks, nsplits=nsplits)

        if len(to_concat_axis_index) > 1:
            raise NotImplementedError

        if to_concat_axis_index:
            axis, output_index = to_concat_axis_index[0]
            indexobj = [slice(None)] * axis + [output_index] + [slice(None)] * (tensor.ndim - axis - 1)

            output_shape = list(tensor.shape)
            output_shape[axis] = len(output_index)
            output_nsplits = list(nsplits)
            output_nsplits[axis] = (output_shape[axis],)
            output_chunks = []
            for idx in itertools.product(*(range(len(it)) for it in (nsplits[:axis]+nsplits[axis+1:]))):
                new_idx = idx[:axis] + (0,) + idx[axis:]
                chunk_idxes = (idx[:axis] + (i,) + idx[axis:]
                               for i in range(len(nsplits[axis])))
                chunks = [tensor.cix[chunk_idx] for chunk_idx in chunk_idxes]
                s = list(chunks[0].shape)
                s[axis] = len(output_index)
                concat_chunk_op = TensorConcatenate(
                    axis=axis, dtype=chunks[0].dtype, sparse=chunks[0].op.sparse)
                concat_chunk = concat_chunk_op.new_chunk(chunks, tuple(s), index=new_idx)
                out_chunk_op = TensorIndex(dtype=concat_chunk.dtype, sparse=concat_chunk.op.sparse)
                out_chunk = out_chunk_op.new_chunk([concat_chunk], tuple(s), indexes=indexobj, index=new_idx)
                output_chunks.append(out_chunk)

            new_op = tensor.op.copy()
            tensor = new_op.new_tensor([op.input], tuple(output_shape), indexes=op.indexes,
                                       chunks=output_chunks, nsplits=output_nsplits)

        return [tensor]


def _getitem(a, item):
    if isinstance(item, (list, tuple)) and \
            all(isinstance(it, slice) and it == slice(None) for it in item):
        # nothing to do
        return a

    # TODO(jisheng): field access, e.g. t['a'], t[['a', 'b']]

    index = process_index(a, item)
    index, shape = get_index_and_shape(a, index)
    op = TensorIndex(dtype=a.dtype, sparse=a.issparse())
    return op(a, index, tuple(shape))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..core import TensorData, Tensor, SparseTensor, ChunkData, Chunk


class TensorOperandMixin(object):
    __slots__ = ()
    _op_module_ = 'tensor'

    def check_inputs(self, inputs):
        pass

    def new_chunks(self, inputs, shape, index=None, output_limit=None, kws=None, dtype=None, **kw):
        output_limit = getattr(self, 'output_limit') if output_limit is None else output_limit

        self.check_inputs(inputs)
        getattr(self, '_set_inputs')(inputs)
        if getattr(self, '_key', None) is None:
            getattr(self, 'update_key')()  # update key when inputs are set

        if isinstance(shape, (list, tuple)) and len(shape) > 0 and isinstance(shape[0], (list, tuple)):
            if len(shape) != output_limit:
                raise ValueError('shape size must be equal to output limit, expect {0}, got {1}'.format(
                    output_limit, len(shape)))
        else:
            shape = [shape] * output_limit

        if kws is not None and kw:
            raise ValueError('can only pass kws or kw')

        chunks = []
        raw_index = index
        for i, s in enumerate(shape):
            dt = None
            if kws:
                kw = kws[i]
                index = kw.pop('index', raw_index)
                dt = kw.pop('dtype', None)
            if dt is None:
                dt = dtype[i] if isinstance(dtype, (tuple, list)) else dtype
            data = ChunkData(_index=index, _shape=s, _op=self, _dtype=dt, **kw)
            chunks.append(Chunk(data))

        setattr(self, 'outputs', chunks)
        return chunks

    def new_tensors(self, inputs, shape, dtype=None, chunks=None, nsplits=None,
                    output_limit=None, kws=None, **kw):
        tensor_cls = SparseTensor if getattr(self, 'issparse')() else Tensor
        output_limit = getattr(self, 'output_limit') if output_limit is None else output_limit

        self.check_inputs(inputs)
        getattr(self, '_set_inputs')(inputs)
        if getattr(self, '_key', None) is None:
            getattr(self, 'update_key')()  # update key when inputs are set

        if isinstance(shape, (list, tuple)) and len(shape) > 0 and isinstance(shape[0], (list, tuple)):
            if not np.isinf(output_limit) and len(shape) != output_limit:
                raise ValueError('shape size must be equal to output limit, expect {0}, got {1}'.format(
                    output_limit, len(shape)))
        else:
            shape = [shape] * output_limit

        if kws is not None and kw:
            raise ValueError('can only pass kws or kw')

        tensors = []
        raw_chunks = chunks
        raw_nsplits = nsplits
        for i, s in enumerate(shape):
            dt = None
            if kws:
                kw = kws[i]
                chunks = kw.pop('chunks', raw_chunks)
                nsplits = kw.pop('nsplits', raw_nsplits)
                dt = kw.pop('dtype', None)
            if nsplits is not None:
                kw['_nsplits'] = nsplits
            if dt is None:
                dt = dtype[i] if isinstance(dtype, (tuple, list)) else dtype
            data = TensorData(_shape=s, _dtype=dt, _op=self,
                              _chunks=chunks, **kw)
            tensors.append(tensor_cls(data))

        setattr(self, 'outputs', tensors)
        if len(tensors) > 1:
            # for each output tensor, hold the reference to the other outputs
            # so that either no one or everyone are gc collected
            for i, t in enumerate(tensors):
                t.data._siblings = [tensor.data for tensor in tensors[:i] + tensors[i+1:]]
        return tensors

    def new_chunk(self, inputs, shape, index=None, **kw):
        if getattr(self, 'output_limit') != 1:
            raise TypeError('cannot new chunk with more than 1 outputs')

        return self.new_chunks(inputs, shape, index=index, **kw)[0]

    def new_tensor(self, inputs, shape, dtype=None, **kw):
        if getattr(self, 'output_limit') != 1:
            raise TypeError('cannot new chunk with more than 1 outputs')

        return self.new_tensors(inputs, shape, dtype=dtype, **kw)[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorBinOp, TensorConstant


class TensorLogAddExp2(operands.LogAddExp2, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLogAddExp2, self).__init__(_casting=casting, _err=err,
                                               _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False

    @classmethod
    def constant_cls(cls):
        return TensorLAE2Constant


class TensorLAE2Constant(operands.LAE2Constant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLAE2Constant, self).__init__(_casting=casting, _err=err,
                                                 _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False


@infer_dtype(np.logaddexp2)
def logaddexp2(x1, x2, out=None, where=None, **kwargs):
    """
    Logarithm of the sum of exponentiations of the inputs in base-2.

    Calculates ``log2(2**x1 + 2**x2)``. This function is useful in machine
    learning when the calculated probabilities of events may be so small as
    to exceed the range of normal floating point numbers.  In such cases
    the base-2 logarithm of the calculated probability can be used instead.
    This function allows adding probabilities stored in such a fashion.

    Parameters
    ----------
    x1, x2 : array_like
        Input values.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    result : Tensor
        Base-2 logarithm of ``2**x1 + 2**x2``.

    See Also
    --------
    logaddexp: Logarithm of the sum of exponentiations of the inputs.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> prob1 = mt.log2(1e-50)
    >>> prob2 = mt.log2(2.5e-50)
    >>> prob12 = mt.logaddexp2(prob1, prob2)
    >>> prob1.execute(), prob2.execute(), prob12.execute()
    (-166.09640474436813, -164.77447664948076, -164.28904982231052)
    >>> (2**prob12).execute()
    3.4999999999999914e-50
    """
    op = TensorLogAddExp2(**kwargs)
    return op(x1, x2, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..datasource import tensor as astensor
from .core import TensorReduction


class TensorAny(operands.Any, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, combine_size=None, **kw):
        super(TensorAny, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                        _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorAny, TensorAny, None


def any(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Test whether any tensor element along a given axis evaluates to True.

    Returns single boolean unless `axis` is not ``None``

    Parameters
    ----------
    a : array_like
        Input tensor or object that can be converted to an array.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a logical OR reduction is performed.
        The default (`axis` = `None`) is to perform a logical OR over all
        the dimensions of the input array. `axis` may be negative, in
        which case it counts from the last to the first axis.

        If this is a tuple of ints, a reduction is performed on multiple
        axes, instead of a single axis or all the axes as before.
    out : Tensor, optional
        Alternate output tensor in which to place the result.  It must have
        the same shape as the expected output and its type is preserved
        (e.g., if it is of type float, then it will remain so, returning
        1.0 for True and 0.0 for False, regardless of the type of `a`).
        See `doc.ufuncs` (Section "Output arguments") for details.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input tensor.

        If the default value is passed, then `keepdims` will not be
        passed through to the `any` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    any : bool or Tensor
        A new boolean or `Tensor` is returned unless `out` is specified,
        in which case a reference to `out` is returned.

    See Also
    --------
    Tensor.any : equivalent method

    all : Test whether all elements along a given axis evaluate to True.

    Notes
    -----
    Not a Number (NaN), positive infinity and negative infinity evaluate
    to `True` because these are not equal to zero.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.any([[True, False], [True, True]]).execute()
    True

    >>> mt.any([[True, False], [False, False]], axis=0).execute()
    array([ True, False])

    >>> mt.any([-1, 0, 5]).execute()
    True

    >>> mt.any(mt.nan).execute()
    True

    """
    a = astensor(a)
    op = TensorAny(axis=axis, dtype=np.dtype(bool), keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from .... import operands
from ....config import options
from ....compat import irange, izip
from ....operands.random import State
from ..utils import decide_chunk_sizes, random_state_data
from .core import TensorRandomOperandMixin


class TensorMultivariateNormal(operands.MultivariateNormal, TensorRandomOperandMixin):
    __slots__ = '_mean', '_cov', '_size', '_check_valid', '_tol'

    def __init__(self, mean=None, cov=None, size=None, check_valid=None, tol=None,
                 state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorMultivariateNormal, self).__init__(_mean=mean, _cov=cov, _size=size,
                                                       _check_valid=check_valid, _tol=tol,
                                                       _state=state, _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, chunk_size=None):
        N = self._mean.size
        if self._size is None:
            shape = (N,)
        else:
            try:
                shape = tuple(self._size) + (N,)
            except TypeError:
                shape = (self._size, N)

        return self.new_tensor(None, shape, raw_chunk_size=chunk_size)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        chunk_size = tensor.params.raw_chunk_size or options.tensor.chunk_size
        nsplits = decide_chunk_sizes(tensor.shape[:-1], chunk_size, tensor.dtype.itemsize) + ((tensor.shape[-1],),)

        mean_chunk = op.mean.chunks[0] if hasattr(op.mean, 'chunks') else op.mean
        cov_chunk = op.cov.chunks[0] if hasattr(op.cov, 'chunks') else op.cov

        idxes = list(itertools.product(*[irange(len(s)) for s in nsplits]))
        states = random_state_data(len(idxes), op.state.random_state) \
            if op.state is not None else [None] * len(idxes)

        out_chunks = []
        for state, out_idx, shape in izip(states, idxes, itertools.product(*nsplits)):
            state = State(np.random.RandomState(state)) if state is not None else None
            chunk_op = op.copy().reset_key()
            chunk_op._state = state
            chunk_op._size = shape[:-1]
            out_chunk = chunk_op.new_chunk([mean_chunk, cov_chunk], shape, index=out_idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape,
                                  chunks=out_chunks, nsplits=nsplits)


def multivariate_normal(random_state, mean, cov, size=None, check_valid=None, tol=None,
                        chunk_size=None, gpu=None, **kw):
    """
    Draw random samples from a multivariate normal distribution.

    The multivariate normal, multinormal or Gaussian distribution is a
    generalization of the one-dimensional normal distribution to higher
    dimensions.  Such a distribution is specified by its mean and
    covariance matrix.  These parameters are analogous to the mean
    (average or "center") and variance (standard deviation, or "width,"
    squared) of the one-dimensional normal distribution.

    Parameters
    ----------
    mean : 1-D array_like, of length N
        Mean of the N-dimensional distribution.
    cov : 2-D array_like, of shape (N, N)
        Covariance matrix of the distribution. It must be symmetric and
        positive-semidefinite for proper sampling.
    size : int or tuple of ints, optional
        Given a shape of, for example, ``(m,n,k)``, ``m*n*k`` samples are
        generated, and packed in an `m`-by-`n`-by-`k` arrangement.  Because
        each sample is `N`-dimensional, the output shape is ``(m,n,k,N)``.
        If no shape is specified, a single (`N`-D) sample is returned.
    check_valid : { 'warn', 'raise', 'ignore' }, optional
        Behavior when the covariance matrix is not positive semidefinite.
    tol : float, optional
        Tolerance when checking the singular values in covariance matrix.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor
        The drawn samples, of shape *size*, if that was provided.  If not,
        the shape is ``(N,)``.

        In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
        value drawn from the distribution.

    Notes
    -----
    The mean is a coordinate in N-dimensional space, which represents the
    location where samples are most likely to be generated.  This is
    analogous to the peak of the bell curve for the one-dimensional or
    univariate normal distribution.

    Covariance indicates the level to which two variables vary together.
    From the multivariate normal distribution, we draw N-dimensional
    samples, :math:`X = [x_1, x_2, ... x_N]`.  The covariance matrix
    element :math:`C_{ij}` is the covariance of :math:`x_i` and :math:`x_j`.
    The element :math:`C_{ii}` is the variance of :math:`x_i` (i.e. its
    "spread").

    Instead of specifying the full covariance matrix, popular
    approximations include:

      - Spherical covariance (`cov` is a multiple of the identity matrix)
      - Diagonal covariance (`cov` has non-negative elements, and only on
        the diagonal)

    This geometrical property can be seen in two dimensions by plotting
    generated data-points:

    >>> mean = [0, 0]
    >>> cov = [[1, 0], [0, 100]]  # diagonal covariance

    Diagonal covariance means that points are oriented along x or y-axis:

    >>> import matplotlib.pyplot as plt
    >>> import mars.tensor as mt
    >>> x, y = mt.random.multivariate_normal(mean, cov, 5000).T
    >>> plt.plot(x.execute(), y.execute(), 'x')
    >>> plt.axis('equal')
    >>> plt.show()

    Note that the covariance matrix must be positive semidefinite (a.k.a.
    nonnegative-definite). Otherwise, the behavior of this method is
    undefined and backwards compatibility is not guaranteed.

    References
    ----------
    .. [1] Papoulis, A., "Probability, Random Variables, and Stochastic
           Processes," 3rd ed., New York: McGraw-Hill, 1991.
    .. [2] Duda, R. O., Hart, P. E., and Stork, D. G., "Pattern
           Classification," 2nd ed., New York: Wiley, 2001.

    Examples
    --------
    >>> mean = (1, 2)
    >>> cov = [[1, 0], [0, 1]]
    >>> x = mt.random.multivariate_normal(mean, cov, (3, 3))
    >>> x.shape
    (3, 3, 2)

    The following is probably true, given that 0.6 is roughly twice the
    standard deviation:

    >>> list(((x[0,0,:] - mean) < 0.6).execute())
    [True, True]
    """
    mean = np.asarray(mean)
    cov = np.asarray(cov)

    if mean.ndim != 1:
        raise ValueError('mean must be 1 dimensional')
    if cov.ndim != 2:
        raise ValueError('cov must be 1 dimensional')
    if len(set(mean.shape + cov.shape)) != 1:
        raise ValueError('mean and cov must have same length')

    if 'dtype' not in kw:
        small_kw = {}
        if check_valid:
            small_kw['check_valid'] = check_valid
        if tol:
            small_kw['tol'] = tol
        kw['dtype'] = np.random.multivariate_normal(mean, cov, size=(0,), **small_kw).dtype

    size = random_state._handle_size(size)
    op = TensorMultivariateNormal(mean=mean, cov=cov, size=size, check_valid=check_valid,
                                  tol=tol, state=random_state._state, gpu=gpu, **kw)
    return op(chunk_size=chunk_size)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from collections import defaultdict

from .kvstore import KVStoreActor
from .utils import SchedulerActor
from ..compat import PY27, OrderedDict3

logger = logging.getLogger(__name__)

_META_CACHE_SIZE = 1000


class WorkerMeta(object):
    __slots__ = 'chunk_size', 'workers'

    def __init__(self, chunk_size=0, workers=None):
        self.chunk_size = chunk_size
        self.workers = workers or ()

    def __eq__(self, other):
        return all(getattr(self, attr) == getattr(other, attr)
                   for attr in self.__slots__)

    def __iter__(self):
        for slot in self.__slots__:
            yield getattr(self, slot)

    def __repr__(self):
        return 'WorkerMeta(%r, %r)' % (self.chunk_size, self.workers)

    if PY27:
        def __getstate__(self):
            return tuple(getattr(self, s) for s in self.__slots__)

        def __setstate__(self, state):
            self.__init__(**dict(zip(self.__slots__, state)))


class ChunkMetaStore(object):
    """
    Storage of chunk meta, holding worker -> chunk relation as well
    """
    def __init__(self):
        self._chunk_metas = dict()
        self._worker_to_chunk_keys = defaultdict(set)

    def __contains__(self, chunk_key):
        return chunk_key in self._chunk_metas

    def __getitem__(self, chunk_key):
        return self._chunk_metas[chunk_key]

    def __setitem__(self, chunk_key, worker_meta):
        if chunk_key in self._chunk_metas:
            self._del_chunk_key_from_workers(chunk_key, self._chunk_metas[chunk_key].workers)

        self._chunk_metas[chunk_key] = worker_meta

        worker_chunks = self._worker_to_chunk_keys
        for w in worker_meta.workers:
            worker_chunks[w].add(chunk_key)

    def __delitem__(self, chunk_key):
        self._del_chunk_key_from_workers(chunk_key, self._chunk_metas[chunk_key].workers)
        del self._chunk_metas[chunk_key]

    def _del_chunk_key_from_workers(self, chunk_key, workers):
        """
        Delete chunk keys from worker
        """
        worker_to_chunks = self._worker_to_chunk_keys
        for w in workers:
            worker_to_chunks[w].remove(chunk_key)

    def get(self, chunk_key, default=None):
        return self._chunk_metas.get(chunk_key, default)

    def get_worker_chunk_keys(self, worker, default=None):
        """
        Get chunk keys held in a worker
        :param worker: worker endpoint
        :param default: default value
        """
        return self._worker_to_chunk_keys.get(worker, default)

    def remove_worker_keys(self, worker, filter_fun=None):
        """
        Remove a worker from storage and return keys of lost chunks
        :param worker: worker endpoint
        :param filter_fun: key filter
        :return: keys of lost chunks
        """
        if worker not in self._worker_to_chunk_keys:
            return []

        filter_fun = filter_fun or (lambda k: True)
        store = self._chunk_metas
        affected = []
        for ckey in tuple(self._worker_to_chunk_keys[worker]):
            if not filter_fun(ckey):
                continue
            self._del_chunk_key_from_workers(ckey, (worker,))
            meta = store[ckey]
            new_workers = meta.workers = tuple(w for w in meta.workers if w != worker)
            if not new_workers:
                affected.append(ckey)
        for ckey in affected:
            del self[ckey]
        if not self._worker_to_chunk_keys[worker]:
            del self._worker_to_chunk_keys[worker]
        return affected


class ChunkMetaCache(ChunkMetaStore):
    """
    Cache of chunk meta with an LRU
    """
    def __init__(self, limit=_META_CACHE_SIZE):
        super(ChunkMetaCache, self).__init__()
        self._chunk_metas = OrderedDict3()
        self._limit = limit

    def __getitem__(self, item):
        self._chunk_metas.move_to_end(item)
        return super(ChunkMetaCache, self).__getitem__(item)

    def get(self, key, default=None):
        try:
            self._chunk_metas.move_to_end(key)
        except KeyError:
            pass
        return super(ChunkMetaCache, self).get(key, default)

    def __setitem__(self, key, value):
        limit = self._limit
        store = self._chunk_metas
        if key in store:
            super(ChunkMetaCache, self).__setitem__(key, value)
            store.move_to_end(key)
        else:
            super(ChunkMetaCache, self).__setitem__(key, value)
            while len(store) > limit:
                dkey, ditem = store.popitem(False)
                self._del_chunk_key_from_workers(dkey, ditem.workers)


class LocalChunkMetaActor(SchedulerActor):
    """
    Actor storing chunk metas and chunk cache
    """
    @classmethod
    def default_name(cls):
        return 's:' + cls.__name__

    def __init__(self, chunk_info_uid=None):
        super(LocalChunkMetaActor, self).__init__()
        self._meta_store = ChunkMetaStore()
        self._meta_broadcasts = dict()
        self._meta_cache = ChunkMetaCache()
        self._chunk_info_uid = chunk_info_uid

        self._kv_store_ref = None

    def post_create(self):
        logger.debug('Actor %s running in process %d at %s', self.uid, os.getpid(), self.address)
        super(LocalChunkMetaActor, self).post_create()

        self._kv_store_ref = self.ctx.actor_ref(KVStoreActor.default_name())
        if not self.ctx.has_actor(self._kv_store_ref):
            self._kv_store_ref = None

    def set_chunk_broadcasts(self, session_id, chunk_key, broadcast_dests):
        """
        Configure broadcast destination addresses. After configuration,
        when the meta of the chunk updates, the update will be broadcast
        into the configured destinations to reduce RPC cost in future.
        :param session_id: session id
        :param chunk_key: chunk key
        :param broadcast_dests: addresses of broadcast destinations, in tuple
        """
        self._meta_broadcasts[(session_id, chunk_key)] = \
            [d for d in broadcast_dests if d != self.address]

    def get_chunk_broadcasts(self, session_id, chunk_key):
        """
        Get chunk broadcast addresses, for test only
        :param session_id: session id
        :param chunk_key: chunk key
        """
        return self._meta_broadcasts.get((session_id, chunk_key))

    def set_chunk_meta(self, session_id, chunk_key, size=None, workers=None):
        """
        Update chunk meta in current storage
        :param session_id: session id
        :param chunk_key: chunk key
        :param size: size of the chunk
        :param workers: workers holding the chunk
        """
        query_key = (session_id, chunk_key)
        # update input with existing value
        if query_key in self._meta_store:
            stored_size, stored_workers = self._meta_store[query_key]
            size = size if size is not None else stored_size
            if workers is not None:
                workers = set(tuple(workers) + stored_workers)

        # sync to external kv store
        if self._kv_store_ref is not None:
            path = '/sessions/%s/chunks/%s' % (session_id, chunk_key)
            if size is not None:
                self._kv_store_ref.write(path + '/data_size', size, _tell=True, _wait=False)
            if workers is not None:
                for w in workers:
                    self._kv_store_ref.write(path + '/workers/%s' % w, '', _tell=True, _wait=False)

        meta = self._meta_store[query_key] = WorkerMeta(size, tuple(workers or ()))

        # broadcast into pre-determined destinations
        futures = []
        if query_key in self._meta_broadcasts:
            for dest in self._meta_broadcasts[query_key]:
                futures.append(
                    self.ctx.actor_ref(self.default_name(), address=dest)
                        .cache_chunk_meta(session_id, chunk_key, meta, _wait=False, _tell=True)
                )
            [f.result() for f in futures]

    def cache_chunk_meta(self, session_id, chunk_key, meta):
        """
        Receive updates for caching

        :param session_id: session id
        :param chunk_key: chunk key
        :param meta: meta data
        """
        query_key = (session_id, chunk_key)
        self._meta_cache[query_key] = meta

    def get_chunk_meta(self, session_id, chunk_key):
        """
        Obtain metadata. If not exists, None will be returned.
        """
        query_key = (session_id, chunk_key)
        try:
            return self._meta_store[query_key]
        except KeyError:
            return self._meta_cache.get(query_key)

    def batch_get_chunk_meta(self, session_id, chunk_keys):
        """
        Obtain metadata in batch
        """
        return [self.get_chunk_meta(session_id, k) for k in chunk_keys]

    def delete_meta(self, session_id, chunk_key):
        """
        Delete metadata from store and cache
        """
        query_key = (session_id, chunk_key)
        try:
            del self._meta_store[query_key]
            if self._kv_store_ref is not None:
                self._kv_store_ref.delete('/sessions/%s/chunks/%s' % (session_id, chunk_key),
                                          recursive=True, _tell=True, _wait=False)
        except KeyError:
            pass
        try:
            del self._meta_cache[query_key]
        except KeyError:
            pass

        # broadcast deletion into pre-determined destinations
        if query_key in self._meta_broadcasts:
            for dest in self._meta_broadcasts[query_key]:
                self.ctx.actor_ref(self.default_name(), address=dest) \
                    .delete_meta(session_id, chunk_key, _wait=False, _tell=True)
            del self._meta_broadcasts[query_key]

    def batch_delete_meta(self, session_id, chunk_keys):
        """
        Delete metadata in batch from store and cache
        """
        for chunk_key in chunk_keys:
            self.delete_meta(session_id, chunk_key)

    def remove_workers_in_session(self, session_id, workers):
        """
        Remove workers from storage given session id and return keys of lost chunks
        :param session_id: session id
        :param workers: worker endpoints
        :return: keys of lost chunks
        """
        logger.debug('Removing workers %r from store', workers)
        removed_chunks = set()
        for w in workers:
            self._meta_cache.remove_worker_keys(w, lambda k: k[0] == session_id)
            removed_chunks.update(self._meta_store.remove_worker_keys(w, lambda k: k[0] == session_id))
        for c in removed_chunks:
            try:
                del self._meta_broadcasts[c]
            except KeyError:
                pass
        return [k[1] for k in removed_chunks]


class ChunkMetaActor(SchedulerActor):
    """
    Actor dispatches chunk meta requests to different scheduler hosts
    """
    @classmethod
    def default_name(cls):
        return 's:' + cls.__name__

    def __init__(self):
        super(ChunkMetaActor, self).__init__()
        self._local_meta_store_ref = None

    def post_create(self):
        logger.debug('Actor %s running in process %d at %s', self.uid, os.getpid(), self.address)
        super(ChunkMetaActor, self).post_create()

        self.set_cluster_info_ref()
        self._local_meta_store_ref = self.ctx.create_actor(
            LocalChunkMetaActor, uid=LocalChunkMetaActor.default_name())

    def set_chunk_broadcasts(self, session_id, chunk_key, broadcast_dests):
        """
        Update metadata broadcast destinations for chunks. After configuration,
        when the meta of the chunk updates, the update will be broadcast
        into the configured destinations to reduce RPC cost in future.
        :param session_id: session id
        :param chunk_key: chunk key
        :param broadcast_dests: destination addresses for broadcast
        """
        addr = self.get_scheduler((session_id, chunk_key))
        self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr) \
            .set_chunk_broadcasts(session_id, chunk_key, broadcast_dests, _tell=True, _wait=False)

    def set_chunk_meta(self, session_id, chunk_key, size=None, workers=None):
        """
        Update chunk metadata
        :param session_id: session id
        :param chunk_key: chunk key
        :param size: size of the chunk
        :param workers: workers holding the chunk
        """
        addr = self.get_scheduler((session_id, chunk_key))
        self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr) \
            .set_chunk_meta(session_id, chunk_key, size=size, workers=workers)

    def set_chunk_size(self, session_id, chunk_key, size):
        self.set_chunk_meta(session_id, chunk_key, size=size)

    def get_chunk_size(self, session_id, chunk_key):
        meta = self.get_chunk_meta(session_id, chunk_key)
        return meta.chunk_size if meta is not None else None

    def add_worker(self, session_id, chunk_key, worker_addr):
        self.set_chunk_meta(session_id, chunk_key, workers=(worker_addr,))

    def get_workers(self, session_id, chunk_key):
        meta = self.get_chunk_meta(session_id, chunk_key)
        return meta.workers if meta is not None else None

    def batch_get_workers(self, session_id, chunk_keys):
        metas = self.batch_get_chunk_meta(session_id, chunk_keys)
        return [meta.workers if meta is not None else None for meta in metas]

    def batch_get_chunk_size(self, session_id, chunk_keys):
        metas = self.batch_get_chunk_meta(session_id, chunk_keys)
        return [meta.chunk_size if meta is not None else None for meta in metas]

    def get_chunk_meta(self, session_id, chunk_key):
        """
        Obtain chunk metadata
        :param session_id: session id
        :param chunk_key: chunk key
        """
        local_result = self._local_meta_store_ref.get_chunk_meta(session_id, chunk_key)
        if local_result is not None:
            return local_result

        addr = self.get_scheduler((session_id, chunk_key))
        meta = self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr) \
            .get_chunk_meta(session_id, chunk_key)
        return meta

    def batch_get_chunk_meta(self, session_id, chunk_keys):
        """
        Obtain chunk metadata in batch
        :param session_id: session id
        :param chunk_keys: chunk keys
        """
        chunk_keys = tuple(chunk_keys)
        query_dict = defaultdict(set)
        meta_dict = dict()

        # try obtaining metadata from local cache
        local_results = self._local_meta_store_ref.batch_get_chunk_meta(session_id, chunk_keys)

        # collect dispatch destinations for non-local metadata
        for chunk_key, local_result in zip(chunk_keys, local_results):
            if local_result is not None:
                meta_dict[chunk_key] = local_result
            else:
                k = (session_id, chunk_key)
                query_dict[self.get_scheduler(k)].add(chunk_key)

        query_dict = dict((k, list(v)) for k, v in query_dict.items())

        # dispatch query
        futures = []
        for addr, keys in query_dict.items():
            futures.append(
                self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr)
                    .batch_get_chunk_meta(session_id, keys, _wait=False)
            )
        # accept results and merge
        for keys, future in zip(query_dict.values(), futures):
            results = future.result()
            meta_dict.update(zip(keys, results))
        return [meta_dict.get(k) for k in chunk_keys]

    def delete_meta(self, session_id, chunk_key):
        query_key = (session_id, chunk_key)
        addr = self.get_scheduler(query_key)
        self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr) \
            .delete_meta(session_id, chunk_key, _tell=True)

    def batch_delete_meta(self, session_id, chunk_keys):
        """
        Delete chunk metadata in batch
        :param session_id: session id
        :param chunk_keys: chunk keys
        """
        # collect dispatch destinations
        chunk_keys = tuple(chunk_keys)
        query_dict = defaultdict(set)
        for chunk_key in chunk_keys:
            k = (session_id, chunk_key)
            query_dict[self.get_scheduler(k)].add(chunk_key)
        # dispatch delete requests and wait
        futures = []
        for addr, keys in query_dict.items():
            futures.append(
                self.ctx.actor_ref(LocalChunkMetaActor.default_name(), address=addr)
                    .batch_delete_meta(session_id, list(keys), _wait=False, _tell=True)
            )
        [f.result() for f in futures]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....operands import Transpose
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


def _reorder(x, axes):
    if x is None:
        return
    return type(x)(np.array(x)[list(axes)].tolist())


class TensorTranspose(Transpose, TensorOperandMixin):
    def __init__(self, axes=None, dtype=None, sparse=False, **kw):
        super(TensorTranspose, self).__init__(_axes=axes, _dtype=dtype,
                                              _sparse=sparse, **kw)

    def __call__(self, a):
        shape = _reorder(a.shape, self._axes)
        return self.new_tensor([a], shape)

    def _set_inputs(self, inputs):
        super(TensorTranspose, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        out_chunks = []
        for c in op.inputs[0].chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = _reorder(c.shape, op.axes)
            chunk_idx = _reorder(c.index, op.axes)
            out_chunk = chunk_op.new_chunk([c], chunk_shape, index=chunk_idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = _reorder(op.inputs[0].nsplits, op.axes)
        return new_op.new_tensors(op.inputs, op.outputs[0].shape,
                                  chunks=out_chunks, nsplits=nsplits)


def transpose(a, axes=None):
    """
    Permute the dimensions of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axes : list of ints, optional
        By default, reverse the dimensions, otherwise permute the axes
        according to the values given.

    Returns
    -------
    p : Tensor
        `a` with its axes permuted.  A view is returned whenever
        possible.

    See Also
    --------
    moveaxis
    argsort

    Notes
    -----
    Use `transpose(a, argsort(axes))` to invert the transposition of tensors
    when using the `axes` keyword argument.

    Transposing a 1-D array returns an unchanged view of the original tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(4).reshape((2,2))
    >>> x.execute()
    array([[0, 1],
           [2, 3]])

    >>> mt.transpose(x).execute()
    array([[0, 2],
           [1, 3]])

    >>> x = mt.ones((1, 2, 3))
    >>> mt.transpose(x, (1, 0, 2)).shape
    (2, 1, 3)

    """
    a = astensor(a)
    if axes:
        if len(axes) != a.ndim:
            raise ValueError("axes don't match tensor")

    axes = axes or list(range(a.ndim))[::-1]
    op = TensorTranspose(axes, dtype=a.dtype, sparse=a.issparse())
    return op(a)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .. import resource
from ..base_app import BaseApplication
from .distributor import SchedulerDistributor
from .service import SchedulerService

logger = logging.getLogger(__name__)


class SchedulerApplication(BaseApplication, SchedulerService):
    service_description = 'Mars Scheduler'
    service_logger = logger

    def config_args(self, parser):
        parser.add_argument('--nproc', help='number of processes')

    def create_pool(self, *args, **kwargs):
        self.n_process = int(self.args.nproc or resource.cpu_count())
        kwargs['distributor'] = SchedulerDistributor(self.n_process)
        return super(SchedulerApplication, self).create_pool(*args, **kwargs)

    def start_service(self):
        super(SchedulerApplication, self).start(self.endpoint, self.args.schedulers, self.pool)

    def stop_service(self):
        super(SchedulerApplication, self).stop(self.pool)


main = SchedulerApplication()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .array import array_module, device
from ...lib import sparse
from ...lib.sparse.core import get_sparse_module, get_array_module, cps, sps, naked
from ...lib.sparse import SparseNDArray
from ..expressions import datasource

logger = logging.getLogger(__name__)


def _create_array(op):
    xp = array_module(op.gpu)

    def inner(func, *args, **kwargs):
        with device(op.device):
            return getattr(xp, func)(*args, **kwargs)

    return inner


def _tensor_ones(ctx, chunk):
    ctx[chunk.key] = _create_array(chunk.op)('ones', chunk.shape, dtype=chunk.op.dtype)


def _tensor_zeros(ctx, chunk):
    if chunk.issparse():
        ctx[chunk.key] = sparse.zeros(chunk.shape, dtype=chunk.op.dtype, gpu=chunk.op.gpu)
    else:
        ctx[chunk.key] = _create_array(chunk.op)('zeros', chunk.shape, dtype=chunk.op.dtype)


def _tensor_empty(ctx, chunk):
    ctx[chunk.key] = _create_array(chunk.op)('empty', chunk.shape, dtype=chunk.op.dtype)


def _tensor_full(ctx, chunk):
    ctx[chunk.key] = _create_array(chunk.op)('full', chunk.shape,
                                             chunk.op.fill_value, dtype=chunk.op.dtype)


def _tensor_ones_like(ctx, chunk):
    if chunk.issparse():
        in_data = naked(ctx[chunk.inputs[0].key])
        xps = get_sparse_module(in_data)
        xp = get_array_module(in_data)
        ctx[chunk.key] = SparseNDArray(xps.csr_matrix(
            (xp.ones_like(in_data.data, dtype=chunk.op.dtype),
             in_data.indices, in_data.indptr), shape=in_data.shape
        ))
    else:
        ctx[chunk.key] = _create_array(chunk.op)(
            'ones_like', ctx[chunk.inputs[0].key], dtype=chunk.op.dtype)


def _tensor_zeros_like(ctx, chunk):
    if chunk.issparse():
        in_data = naked(ctx[chunk.inputs[0].key])
        xps = get_sparse_module(in_data)
        xp = get_array_module(in_data)
        ctx[chunk.key] = SparseNDArray(xps.csr_matrix(
            (xp.zeros_like(in_data.data, dtype=chunk.op.dtype),
             in_data.indices, in_data.indptr), shape=in_data.shape
        ))
    else:
        ctx[chunk.key] = _create_array(chunk.op)(
            'zeros_like', ctx[chunk.inputs[0].key], dtype=chunk.op.dtype)


def _tensor_empty_like(ctx, chunk):
    if chunk.issparse():
        in_data = naked(ctx[chunk.inputs[0].key])
        xps = get_array_module(in_data)
        xp = get_array_module(in_data)
        ctx[chunk.key] = SparseNDArray(xps.csr_matrix(
            (xp.empty_like(in_data.data, dtype=chunk.op.dtype),
             in_data.indices, in_data.indptr), shape=in_data.shape
        ))
    else:
        ctx[chunk.key] = _create_array(chunk.op)(
            'empty_like', ctx[chunk.inputs[0].key], dtype=chunk.op.dtype)


def _tensor_arange(ctx, chunk):
    ctx[chunk.key] = _create_array(chunk.op)(
        'arange', chunk.op.start, chunk.op.stop, chunk.op.step, dtype=chunk.op.dtype)


def _tensor_diag(ctx, chunk):
    if chunk.issparse():
        ctx[chunk.key] = sparse.diag(ctx[chunk.inputs[0].key], k=chunk.op.k, gpu=chunk.op.gpu)
    else:
        ctx[chunk.key] = _create_array(chunk.op)(
            'diag', ctx[chunk.inputs[0].key], k=chunk.op.k)


def _tensor_eye(ctx, chunk):
    if chunk.issparse():
        ctx[chunk.key] = sparse.eye(chunk.shape[0], M=chunk.shape[1], k=chunk.op.k,
                                    dtype=chunk.op.dtype, gpu=chunk.op.gpu)
    else:
        ctx[chunk.key] = _create_array(chunk.op)(
            'eye', chunk.shape[0], M=chunk.shape[1], k=chunk.op.k, dtype=chunk.op.dtype)


def _tensor_linspace(ctx, chunk):
    ctx[chunk.key] = _create_array(chunk.op)(
        'linspace', chunk.op.start, chunk.op.stop, num=chunk.op.num,
        endpoint=chunk.op.endpoint, dtype=chunk.op.dtype)


def _tensor_tri(ctx, chunk):
    f = 'triu' if isinstance(chunk.op, datasource.TensorTriu) else 'tril'
    if chunk.op.sparse:
        ctx[chunk.key] = getattr(sparse, f)(ctx[chunk.inputs[0].key], k=chunk.op.k)
    else:
        ctx[chunk.key] = _create_array(chunk.op)(
            f, ctx[chunk.inputs[0].key], chunk.op.k)


def _tensor_array_data_source(ctx, chunk):
    ctx[chunk.key] = array_module(chunk.op.gpu).asarray(chunk.op.data)


def _tensor_csr_matrix_data_source(ctx, chunk):
    xps = cps if chunk.op.gpu else sps
    ctx[chunk.key] = SparseNDArray(xps.csr_matrix(
        (chunk.op.data, chunk.op.indices, chunk.op.indptr), shape=chunk.op.shape
    ))


def _tensor_sparse_to_dense(ctx, chunk):
    ctx[chunk.key] = ctx[chunk.inputs[0].key].toarray()


def _tensor_dense_to_sparse(ctx, chunk):
    in_data = naked(ctx[chunk.inputs[0].key])
    xps = cps if chunk.op.gpu else sps
    ctx[chunk.key] = SparseNDArray(xps.csr_matrix(in_data))


def _tensor_fetch_chunk(ctx, chunk):
    # nothing need to do
    return


def _scalar(ctx, chunk):
    if chunk.ndim != 0:
        raise ValueError('Missing op for chunk')

    ctx[chunk.key] = _create_array(chunk.op)('asarray', chunk.op.data)


def register_data_source_handler():
    from .core import register

    register(datasource.TensorOnes, _tensor_ones)
    register(datasource.TensorZeros, _tensor_zeros)
    register(datasource.TensorEmpty, _tensor_empty)
    register(datasource.TensorFull, _tensor_full)
    register(datasource.TensorOnesLike, _tensor_ones_like)
    register(datasource.TensorZerosLike, _tensor_zeros_like)
    register(datasource.TensorEmptyLike, _tensor_empty_like)
    register(datasource.TensorArange, _tensor_arange)
    register(datasource.TensorDiag, _tensor_diag)
    register(datasource.TensorEye, _tensor_eye)
    register(datasource.TensorLinspace, _tensor_linspace)
    register(datasource.TensorTriu, _tensor_tri)
    register(datasource.TensorTril, _tensor_tri)
    register(datasource.ArrayDataSource, _tensor_array_data_source)
    register(datasource.CSRMatrixDataSource, _tensor_csr_matrix_data_source)
    register(datasource.SparseToDense, _tensor_sparse_to_dense)
    register(datasource.DenseToSparse, _tensor_dense_to_sparse)
    register(datasource.TensorFetchChunk, _tensor_fetch_chunk)
    register(datasource.Scalar, _scalar)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorGumbel(operands.Gumbel, TensorRandomOperandMixin):
    __slots__ = '_loc', '_scale', '_size'
    _input_fields_ = ['_loc', '_scale']

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorGumbel, self).__init__(_state=state, _size=size, _dtype=dtype,
                                           _gpu=gpu, **kw)

    def __call__(self, loc, scale, chunk_size=None):
        return self.new_tensor([loc, scale], None, raw_chunk_size=chunk_size)


def gumbel(random_state, loc=0.0, scale=1.0, size=None, chunk_size=None, gpu=None, **kw):
    r"""
    Draw samples from a Gumbel distribution.

    Draw samples from a Gumbel distribution with specified location and
    scale.  For more information on the Gumbel distribution, see
    Notes and References below.

    Parameters
    ----------
    loc : float or array_like of floats, optional
        The location of the mode of the distribution. Default is 0.
    scale : float or array_like of floats, optional
        The scale parameter of the distribution. Default is 1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``loc`` and ``scale`` are both scalars.
        Otherwise, ``np.broadcast(loc, scale).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized Gumbel distribution.

    See Also
    --------
    scipy.stats.gumbel_l
    scipy.stats.gumbel_r
    scipy.stats.genextreme
    weibull

    Notes
    -----
    The Gumbel (or Smallest Extreme Value (SEV) or the Smallest Extreme
    Value Type I) distribution is one of a class of Generalized Extreme
    Value (GEV) distributions used in modeling extreme value problems.
    The Gumbel is a special case of the Extreme Value Type I distribution
    for maximums from distributions with "exponential-like" tails.

    The probability density for the Gumbel distribution is

    .. math:: p(x) = \frac{e^{-(x - \mu)/ \beta}}{\beta} e^{ -e^{-(x - \mu)/
              \beta}},

    where :math:`\mu` is the mode, a location parameter, and
    :math:`\beta` is the scale parameter.

    The Gumbel (named for German mathematician Emil Julius Gumbel) was used
    very early in the hydrology literature, for modeling the occurrence of
    flood events. It is also used for modeling maximum wind speed and
    rainfall rates.  It is a "fat-tailed" distribution - the probability of
    an event in the tail of the distribution is larger than if one used a
    Gaussian, hence the surprisingly frequent occurrence of 100-year
    floods. Floods were initially modeled as a Gaussian process, which
    underestimated the frequency of extreme events.

    It is one of a class of extreme value distributions, the Generalized
    Extreme Value (GEV) distributions, which also includes the Weibull and
    Frechet.

    The function has a mean of :math:`\mu + 0.57721\beta` and a variance
    of :math:`\frac{\pi^2}{6}\beta^2`.

    References
    ----------
    .. [1] Gumbel, E. J., "Statistics of Extremes,"
           New York: Columbia University Press, 1958.
    .. [2] Reiss, R.-D. and Thomas, M., "Statistical Analysis of Extreme
           Values from Insurance, Finance, Hydrology and Other Fields,"
           Basel: Birkhauser Verlag, 2001.

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> mu, beta = 0, 0.1 # location and scale
    >>> s = mt.random.gumbel(mu, beta, 1000).execute()

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> count, bins, ignored = plt.hist(s, 30, normed=True)
    >>> plt.plot(bins, (1/beta)*np.exp(-(bins - mu)/beta)
    ...          * np.exp( -np.exp( -(bins - mu) /beta) ),
    ...          linewidth=2, color='r')
    >>> plt.show()

    Show how an extreme value distribution can arise from a Gaussian process
    and compare to a Gaussian:

    >>> means = []
    >>> maxima = []
    >>> for i in range(0,1000) :
    ...    a = mt.random.normal(mu, beta, 1000)
    ...    means.append(a.mean().execute())
    ...    maxima.append(a.max().execute())
    >>> count, bins, ignored = plt.hist(maxima, 30, normed=True)
    >>> beta = mt.std(maxima) * mt.sqrt(6) / mt.pi
    >>> mu = mt.mean(maxima) - 0.57721*beta
    >>> plt.plot(bins, ((1/beta)*mt.exp(-(bins - mu)/beta)
    ...          * mt.exp(-mt.exp(-(bins - mu)/beta))).execute(),
    ...          linewidth=2, color='r')
    >>> plt.plot(bins, (1/(beta * mt.sqrt(2 * mt.pi))
    ...          * mt.exp(-(bins - mu)**2 / (2 * beta**2))).execute(),
    ...          linewidth=2, color='g')
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().gumbel(
            handle_array(loc), handle_array(scale), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorGumbel(state=random_state._state, size=size, gpu=gpu, **kw)
    return op(loc, scale, chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorBinOp, TensorConstant


class TensorArctan2(operands.Arctan2, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorArctan2, self).__init__(_casting=casting, _err=err,
                                            _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        # x2 is sparse or not does not matter
        return x1.issparse()

    @classmethod
    def constant_cls(cls):
        return TensorArct2Constant


class TensorArct2Constant(operands.Arct2Constant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorArct2Constant, self).__init__(_casting=casting, _err=err,
                                                  _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse() and np.isscalar(x2):
            return True
        elif x1 == 0:
            return True
        return False


@infer_dtype(np.arctan2)
def arctan2(x1, x2, out=None, where=None, **kwargs):
    """
    Element-wise arc tangent of ``x1/x2`` choosing the quadrant correctly.

    The quadrant (i.e., branch) is chosen so that ``arctan2(x1, x2)`` is
    the signed angle in radians between the ray ending at the origin and
    passing through the point (1,0), and the ray ending at the origin and
    passing through the point (`x2`, `x1`).  (Note the role reversal: the
    "`y`-coordinate" is the first function parameter, the "`x`-coordinate"
    is the second.)  By IEEE convention, this function is defined for
    `x2` = +/-0 and for either or both of `x1` and `x2` = +/-inf (see
    Notes for specific values).

    This function is not defined for complex-valued arguments; for the
    so-called argument of complex values, use `angle`.

    Parameters
    ----------
    x1 : array_like, real-valued
        `y`-coordinates.
    x2 : array_like, real-valued
        `x`-coordinates. `x2` must be broadcastable to match the shape of
        `x1` or vice versa.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    angle : Tensor
        Array of angles in radians, in the range ``[-pi, pi]``.

    See Also
    --------
    arctan, tan, angle

    Notes
    -----
    *arctan2* is identical to the `atan2` function of the underlying
    C library.  The following special values are defined in the C
    standard: [1]_

    ====== ====== ================
    `x1`   `x2`   `arctan2(x1,x2)`
    ====== ====== ================
    +/- 0  +0     +/- 0
    +/- 0  -0     +/- pi
     > 0   +/-inf +0 / +pi
     < 0   +/-inf -0 / -pi
    +/-inf +inf   +/- (pi/4)
    +/-inf -inf   +/- (3*pi/4)
    ====== ====== ================

    Note that +0 and -0 are distinct floating point numbers, as are +inf
    and -inf.

    References
    ----------
    .. [1] ISO/IEC standard 9899:1999, "Programming language C."

    Examples
    --------
    Consider four points in different quadrants:
    >>> import mars.tensor as mt

    >>> x = mt.array([-1, +1, +1, -1])
    >>> y = mt.array([-1, -1, +1, +1])
    >>> (mt.arctan2(y, x) * 180 / mt.pi).execute()
    array([-135.,  -45.,   45.,  135.])

    Note the order of the parameters. `arctan2` is defined also when `x2` = 0
    and at several other special points, obtaining values in
    the range ``[-pi, pi]``:

    >>> mt.arctan2([1., -1.], [0., 0.]).execute()
    array([ 1.57079633, -1.57079633])
    >>> mt.arctan2([0., 0., mt.inf], [+0., -0., mt.inf]).execute()
    array([ 0.        ,  3.14159265,  0.78539816])
    """
    op = TensorArctan2(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import heapq
import logging
import os
import random
import time
from collections import defaultdict

from .. import promise
from ..config import options
from ..utils import log_unhandled
from .chunkmeta import ChunkMetaActor
from .resource import ResourceActor
from .utils import SchedulerActor

logger = logging.getLogger(__name__)


class ChunkPriorityItem(object):
    """
    Class providing an order for operands for assignment
    """
    def __init__(self, session_id, op_key, op_info, callback):
        self._op_key = op_key
        self._session_id = session_id
        self._op_info = op_info
        self._target_worker = op_info.get('target_worker')
        self._callback = callback

        self._priority = ()
        self.update_priority(op_info['optimize'])

    def update_priority(self, priority_data, copyobj=False):
        obj = self
        if copyobj:
            obj = copy.deepcopy(obj)

        priorities = []
        priorities.extend([
            priority_data.get('depth', 0),
            priority_data.get('demand_depths', ()),
            -priority_data.get('successor_size', 0),
            -priority_data.get('placement_order', 0),
            priority_data.get('descendant_size'),
        ])
        obj._priority = tuple(priorities)
        return obj

    @property
    def session_id(self):
        return self._session_id

    @property
    def op_key(self):
        return self._op_key

    @property
    def target_worker(self):
        return self._target_worker

    @target_worker.setter
    def target_worker(self, value):
        self._target_worker = value

    @property
    def callback(self):
        return self._callback

    @property
    def op_info(self):
        return self._op_info

    def __repr__(self):
        return '<ChunkPriorityItem(%s(%s))>' % (self.op_key, self.op_info['op_name'])

    def __lt__(self, other):
        return self._priority > other._priority


class AssignerActor(SchedulerActor):
    """
    Actor handling worker assignment requests from operands.
    Note that this actor does not assign workers itself.
    """
    @staticmethod
    def gen_name(session_id):
        return 's:assigner$%s' % session_id

    def __init__(self):
        super(AssignerActor, self).__init__()
        self._requests = dict()
        self._req_heap = []

        self._cluster_info_ref = None
        self._actual_ref = None
        self._resource_actor_ref = None

        self._worker_metrics = None
        # since worker metrics does not change frequently, we update it
        # only when it is out of date
        self._worker_metric_time = time.time() - 2

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        self.set_cluster_info_ref()
        # the ref of the actor actually handling assignment work
        self._actual_ref = self.ctx.create_actor(AssignEvaluationActor, self.ref())
        self._resource_actor_ref = self.get_actor_ref(ResourceActor.default_name())

    def allocate_top_resources(self):
        self._actual_ref.allocate_top_resources(_tell=True)

    @promise.reject_on_exception
    @log_unhandled
    def apply_for_resource(self, session_id, op_key, op_info, callback=None):
        """
        Register resource request for an operand
        :param session_id: session id
        :param op_key: operand key
        :param op_info: operand information, should be a dict
        :param callback: promise callback, called when the resource is assigned
        """
        t = time.time()
        if self._worker_metrics is None or self._worker_metric_time + 1 < time.time():
            # update worker metrics from ResourceActor
            self._worker_metrics = self._resource_actor_ref.get_workers_meta()
            self._worker_metric_time = t

        priority_item = ChunkPriorityItem(session_id, op_key, op_info, callback)
        if priority_item.target_worker not in self._worker_metrics:
            priority_item.target_worker = None
        self._requests[op_key] = priority_item
        heapq.heappush(self._req_heap, priority_item)
        self._actual_ref.allocate_top_resources(_tell=True)

    @log_unhandled
    def update_priority(self, op_key, priority_data):
        """
        Update priority data for an operand. The priority item will be
        pushed into priority queue again.
        :param op_key: operand key
        :param priority_data: new priority data
        """
        if op_key not in self._requests:
            return
        obj = self._requests[op_key].update_priority(priority_data, copyobj=True)
        heapq.heappush(self._req_heap, obj)

    @log_unhandled
    def remove_apply(self, op_key):
        """
        Cancel request for an operand
        :param op_key: operand key
        """
        if op_key in self._requests:
            del self._requests[op_key]

    def pop_head(self):
        """
        Pop and obtain top-priority request from queue
        :return: top item
        """
        item = None
        while self._req_heap:
            item = heapq.heappop(self._req_heap)
            if item.op_key in self._requests:
                break
            else:
                item = None
        return item

    def extend(self, items):
        """
        Extend heap by an iterable object. The heap will be reheapified.
        :param items: priority items
        """
        self._req_heap.extend(items)
        heapq.heapify(self._req_heap)


class AssignEvaluationActor(SchedulerActor):
    """
    Actor assigning operands to workers
    """
    def __init__(self, assigner_ref):
        super(AssignEvaluationActor, self).__init__()
        self._worker_metrics = None
        self._worker_metric_time = time.time() - 2

        self._cluster_info_ref = None
        self._assigner_ref = assigner_ref
        self._resource_actor_ref = None
        self._chunk_meta_ref = None

        self._sufficient_operands = set()
        self._operand_sufficient_time = dict()

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        self.set_cluster_info_ref()
        self._assigner_ref = self.ctx.actor_ref(self._assigner_ref)
        self._resource_actor_ref = self.get_actor_ref(ResourceActor.default_name())
        self._chunk_meta_ref = self.ctx.actor_ref(ChunkMetaActor.default_name())

        self.periodical_allocate()

    def periodical_allocate(self):
        self.allocate_top_resources()
        self.ref().periodical_allocate(_tell=True, _delay=1)

    def allocate_top_resources(self):
        """
        Allocate resources given the order in AssignerActor
        """
        t = time.time()
        if self._worker_metrics is None or self._worker_metric_time + 1 < time.time():
            # update worker metrics from ResourceActor
            self._worker_metrics = self._resource_actor_ref.get_workers_meta()
            self._worker_metric_time = t
        if not self._worker_metrics:
            return

        unassigned = []
        reject_workers = set()
        # the assigning procedure will continue till
        while len(reject_workers) < len(self._worker_metrics):
            item = self._assigner_ref.pop_head()
            if not item:
                break

            try:
                alloc_ep, rejects = self._allocate_resource(
                    item.session_id, item.op_key, item.op_info, item.target_worker,
                    reject_workers=reject_workers, callback=item.callback)
            except:
                logger.exception('Unexpected error occurred in %s', self.uid)
                unassigned.append(item)
                continue

            # collect workers failed to assign operand to
            reject_workers.update(rejects)
            if alloc_ep:
                # assign successfully, we remove the application
                self._assigner_ref.remove_apply(item.op_key)
            else:
                # put the unassigned item into unassigned list to add back to the queue later
                unassigned.append(item)
        if unassigned:
            # put unassigned back to the queue, if any
            self._assigner_ref.extend(unassigned)

    @promise.reject_on_exception
    @log_unhandled
    def _allocate_resource(self, session_id, op_key, op_info, target_worker=None, reject_workers=None, callback=None):
        """
        Allocate resource for single operand
        :param session_id: session id
        :param op_key: operand key
        :param op_info: operand info dict
        :param target_worker: worker to allocate, can be None
        :param reject_workers: workers denied to assign to
        :param callback: promise callback from operands
        """
        if target_worker not in self._worker_metrics:
            target_worker = None

        reject_workers = reject_workers or set()

        op_io_meta = op_info['io_meta']
        input_chunk_keys = op_io_meta['input_chunks']
        metas = self._get_chunks_meta(session_id, input_chunk_keys)
        input_sizes = dict((k, meta.chunk_size) for k, meta in metas.items())
        output_size = op_info['output_size']

        if target_worker is None:
            op_name = op_info['op_name']
            who_has = dict((k, meta.workers) for k, meta in metas.items())

            candidate_workers = self._get_eps_by_worker_locality(input_chunk_keys, who_has, input_sizes)
            locality_workers = set(candidate_workers)
            if self._is_stats_sufficient(op_name):
                ep = self._get_ep_by_worker_stats(input_chunk_keys, who_has, input_sizes, output_size, op_name)
                if ep:
                    candidate_workers.append(ep)
        else:
            candidate_workers = [target_worker]
            locality_workers = {target_worker}

        candidate_workers = [w for w in candidate_workers if w not in reject_workers]
        if not candidate_workers:
            return None, []

        # todo make more detailed allocation plans
        # alloc_dict = dict(cpu=1, memory=sum(six.itervalues(input_sizes)) + output_size)
        alloc_dict = dict(cpu=1, memory=output_size)
        rejects = []
        for worker_ep in candidate_workers:
            if self._resource_actor_ref.allocate_resource(session_id, op_key, worker_ep, alloc_dict):
                if worker_ep in locality_workers:
                    logger.debug('Operand %s(%s) allocated to run in %s', op_key, op_info['op_name'], worker_ep)
                else:
                    logger.debug('Operand %s(%s) allocated to run in %s given collected statistics',
                                 op_key, op_info['op_name'], worker_ep)

                self.tell_promise(callback, worker_ep)
                return worker_ep, rejects
            rejects.append(worker_ep)
        return None, rejects

    def _get_chunks_meta(self, session_id, keys):
        return dict(zip(keys, self._chunk_meta_ref.batch_get_chunk_meta(session_id, keys)))

    def _get_op_metric_item(self, ep, op_name, item):
        return self._get_metric_item(ep, 'calc_speed.' + op_name, item)

    def _get_metric_item(self, ep, metric, item):
        return self._worker_metrics[ep].get('stats', {}).get(metric, {}).get(item, 0)

    def _is_stats_sufficient(self, op_name):
        if not options.scheduler.enable_chunk_relocation:
            return False
        if op_name is None:
            return False

        if op_name in self._sufficient_operands:
            return True

        t = time.time()
        if self._operand_sufficient_time.get(op_name, 0) > t - 1:
            return False

        if any(self._worker_metrics[ep].get('stats', {}).get('max_est_finish_time') is None
               for ep in self._worker_metrics):
            self._operand_sufficient_time[op_name] = t
            return False
        minimal_stats = options.optimize.min_stats_count
        minimal_workers = len(self._worker_metrics) * options.optimize.stats_sufficient_ratio - 1e-6
        if sum(self._get_metric_item(ep, 'net_transfer_speed', 'count') >= minimal_stats
               for ep in self._worker_metrics) < minimal_workers:
            if all(self._get_metric_item(ep, 'net_transfer_speed', 'count') == 0 for ep in self._worker_metrics) or \
                    all(self._worker_metrics[ep].get('submitted_count', 0) > 0 for ep in self._worker_metrics):
                self._operand_sufficient_time[op_name] = t
                return False
        if any(self._get_op_metric_item(ep, op_name, 'count') < minimal_stats for ep in self._worker_metrics):
            self._operand_sufficient_time[op_name] = t
            return False
        if any(abs(self._get_op_metric_item(ep, op_name, 'mean')) < 1e-6 for ep in self._worker_metrics):
            self._operand_sufficient_time[op_name] = t
            return False

        if op_name in self._operand_sufficient_time:
            del self._operand_sufficient_time[op_name]
        self._sufficient_operands.add(op_name)
        return True

    def _get_eps_by_worker_locality(self, input_keys, chunk_workers, input_sizes):
        locality_data = defaultdict(lambda: 0)
        for k in input_keys:
            if k in chunk_workers:
                for ep in chunk_workers[k]:
                    locality_data[ep] += input_sizes[k]
        workers = list(self._worker_metrics.keys())
        random.shuffle(workers)
        max_locality = -1
        max_eps = []
        for ep in workers:
            if locality_data[ep] > max_locality:
                max_locality = locality_data[ep]
                max_eps = [ep]
            elif locality_data[ep] == max_locality:
                max_eps.append(ep)
        return max_eps

    def _get_ep_by_worker_stats(self, input_keys, chunk_workers, input_sizes, output_size, op_name):
        ep_net_speeds = dict()
        if any(self._get_metric_item(ep, 'net_transfer_speed', 'count') <= options.optimize.min_stats_count
               for ep in self._worker_metrics):
            sum_speeds, sum_records = 0, 0
            for ep in self._worker_metrics.keys():
                sum_speeds += self._get_metric_item(ep, 'net_transfer_speed', 'mean')
                sum_records += self._get_metric_item(ep, 'net_transfer_speed', 'count')
            avg_speed = sum_speeds * 1.0 / sum_records
        else:
            avg_speed = 0
        for ep in self._worker_metrics:
            if self._get_metric_item(ep, 'net_transfer_speed', 'count') > options.optimize.min_stats_count:
                ep_net_speeds[ep] = self._get_metric_item(ep, 'net_transfer_speed', 'mean')
            else:
                ep_net_speeds[ep] = avg_speed

        start_time = time.time()
        ep_transmit_times = defaultdict(list)
        ep_calc_time = defaultdict(lambda: 0)
        locality_data = defaultdict(lambda: 0)
        for key in input_keys:
            contain_eps = chunk_workers.get(key, set())
            for ep in self._worker_metrics:
                if ep not in contain_eps:
                    ep_transmit_times[ep].append(input_sizes[key] * 1.0 / ep_net_speeds[ep])
                else:
                    locality_data[ep] += input_sizes[key]

        ep_transmit_time = dict()
        for ep in self._worker_metrics:
            if ep_transmit_times[ep]:
                ep_transmit_time[ep] = sum(ep_transmit_times[ep])
            else:
                ep_transmit_time[ep] = 0

        max_eps = []
        max_locality = 0
        for ep, locality in locality_data.items():
            if locality == max_locality:
                max_eps.append(ep)
            elif locality > max_locality:
                max_eps = [ep]
                max_locality = locality
        max_eps = set(max_eps)

        for ep in self._worker_metrics:
            ep_calc_time[ep] = sum(input_sizes.values()) * 1.0 / self._get_op_metric_item(ep, op_name, 'mean')

        min_locality_exec_time = 0
        for ep in max_eps:
            finish_time = max(start_time, self._worker_metrics[ep].get('stats', {}).get('max_est_finish_time', 0.0)) + \
                          ep_transmit_time[ep] + ep_calc_time[ep]
            min_locality_exec_time = max(finish_time, min_locality_exec_time)

        min_finish_time = None
        min_ep = None
        for ep in self._worker_metrics:
            if ep_calc_time[ep] < 2 * ep_transmit_time[ep]:
                continue
            finish_time = max(start_time, self._worker_metrics[ep].get('stats', {}).get('min_est_finish_time', 0.0)) + \
                          ep_transmit_time[ep] + ep_calc_time[ep]
            if ep not in max_eps:
                finish_time += output_size * 1.0 / ep_net_speeds[ep]
            if finish_time - min_locality_exec_time > 1e-6:
                continue
            if min_finish_time is None or min_finish_time > finish_time:
                min_finish_time = finish_time
                min_ep = ep
        return min_ep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import inject_dtype
from .core import TensorCompare, TensorCompareConstant


class TensorLessThan(operands.LessThan, TensorCompare):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=None, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLessThan, self).__init__(_casting=casting, _err=err,
                                             _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def constant_cls(cls):
        return TensorLtConstant


class TensorLtConstant(operands.LtConstant, TensorCompareConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=None, **kw):
        err = err if err is not None else np.geterr()
        super(TensorLtConstant, self).__init__(_casting=casting, _err=err,
                                               _dtype=dtype, _sparse=sparse, **kw)


@inject_dtype(np.bool_)
def less(x1, x2, out=None, where=None, **kwargs):
    """
    Return the truth value of (x1 < x2) element-wise.

    Parameters
    ----------
    x1, x2 : array_like
        Input tensors.  If ``x1.shape != x2.shape``, they must be
        broadcastable to a common shape (which may be the shape of one or
        the other).
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : bool or Tensor of bool
        Array of bools, or a single bool if `x1` and `x2` are scalars.

    See Also
    --------
    greater, less_equal, greater_equal, equal, not_equal

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.less([1, 2], [2, 2]).execute()
    array([ True, False])
    """
    op = TensorLessThan(**kwargs)
    return op(x1, x2, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import inject_dtype
from .core import TensorBinOp, TensorConstant


class TensorIsclose(operands.Isclose, TensorBinOp):
    def __init__(self, rtol=None, atol=None, equal_nan=None,
                 casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorIsclose, self).__init__(_rtol=rtol, _atol=atol, _equal_nan=equal_nan,
                                            _casting=casting, _err=err,
                                            _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False

    @classmethod
    def constant_cls(cls):
        return TensorIscloseConstant

    def to_constant(self, x1, x2):
        constant_op = TensorIscloseConstant(getattr(self, '_rtol'), getattr(self, '_atol'),
                                            getattr(self, '_equal_nan'), getattr(self, '_casting'),
                                            getattr(self, '_err'), getattr(self, '_dtype'),
                                            getattr(self, '_sparse'))
        return constant_op(x1, x2)


class TensorIscloseConstant(operands.IscloseConstant, TensorConstant):
    def __init__(self, rtol=None, atol=None, equal_nan=None,
                 casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorIscloseConstant, self).__init__(_rtol=rtol, _atol=atol, _equal_nan=equal_nan,
                                                    _casting=casting, _err=err,
                                                    _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse() and \
                np.isscalar(x2) and not np.isclose(x2, 0):
            return True
        if hasattr(x2, 'issparse') and x2.issparse() and \
                np.isscalar(x1) and not np.isclose(x1, 0):
            return True
        return False


@inject_dtype(np.bool_)
def isclose(a, b, rtol=1e-05, atol=1e-08, equal_nan=False, **kwargs):
    """
    Returns a boolean tensor where two tensors are element-wise equal within a
    tolerance.

    The tolerance values are positive, typically very small numbers.  The
    relative difference (`rtol` * abs(`b`)) and the absolute difference
    `atol` are added together to compare against the absolute difference
    between `a` and `b`.

    Parameters
    ----------
    a, b : array_like
        Input tensors to compare.
    rtol : float
        The relative tolerance parameter (see Notes).
    atol : float
        The absolute tolerance parameter (see Notes).
    equal_nan : bool
        Whether to compare NaN's as equal.  If True, NaN's in `a` will be
        considered equal to NaN's in `b` in the output tensor.

    Returns
    -------
    y : array_like
        Returns a boolean tensor of where `a` and `b` are equal within the
        given tolerance. If both `a` and `b` are scalars, returns a single
        boolean value.

    See Also
    --------
    allclose

    Notes
    -----

    For finite values, isclose uses the following equation to test whether
    two floating point values are equivalent.

     absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))

    The above equation is not symmetric in `a` and `b`, so that
    `isclose(a, b)` might be different from `isclose(b, a)` in
    some rare cases.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.isclose([1e10,1e-7], [1.00001e10,1e-8]).execute()
    array([True, False])
    >>> mt.isclose([1e10,1e-8], [1.00001e10,1e-9]).execute()
    array([True, True])
    >>> mt.isclose([1e10,1e-8], [1.0001e10,1e-9]).execute()
    array([False, True])
    >>> mt.isclose([1.0, mt.nan], [1.0, mt.nan]).execute()
    array([True, False])
    >>> mt.isclose([1.0, mt.nan], [1.0, mt.nan], equal_nan=True).execute()
    array([True, True])
    """
    op = TensorIsclose(rtol=rtol, atol=atol, equal_nan=equal_nan, **kwargs)
    return op(a, b)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .. import resource
from ..config import options
from ..base_app import BaseApplication
from ..errors import StartArgumentError
from .distributor import WorkerDistributor
from .service import WorkerService

logger = logging.getLogger(__name__)


class WorkerApplication(BaseApplication, WorkerService):
    """
    Main function class of Mars Worker
    """
    service_description = 'Mars Worker'
    service_logger = logger

    def __init__(self):
        super(BaseApplication, self).__init__()
        super(WorkerService, self).__init__()

    def config_args(self, parser):
        parser.add_argument('--cpu-procs', help='number of processes used for cpu')
        parser.add_argument('--io-procs', help='number of processes used for io')
        parser.add_argument('--phy-mem', help='physical memory size limit')
        parser.add_argument('--ignore-avail-mem', action='store_true', help='ignore available memory')
        parser.add_argument('--cache-mem', help='cache memory size limit')
        parser.add_argument('--disk', help='disk size limit')
        parser.add_argument('--spill-dir', help='spill directory')
        parser.add_argument('--plasma-one-mapped-file', action='store_true',
                            help='path of Plasma UNIX socket')

    def validate_arguments(self):
        if not self.args.schedulers and not self.args.kv_store:
            raise StartArgumentError('either schedulers or url of kv store is required.')
        if not self.args.advertise:
            raise StartArgumentError('advertise address is required.')

    def create_pool(self, *args, **kwargs):
        # here we create necessary actors on worker
        # and distribute them over processes
        mem_stats = resource.virtual_memory()

        options.worker.cpu_process_count = int(self.args.cpu_procs or resource.cpu_count())
        options.worker.io_process_count = int(self.args.io_procs or '1')
        options.worker.physical_memory_limit_hard = self._calc_size_limit(
            self.args.phy_mem or options.worker.physical_memory_limit_hard, mem_stats.total
        )
        options.worker.physical_memory_limit_soft = self._calc_size_limit(
            self.args.phy_mem or options.worker.physical_memory_limit_soft or '48%', mem_stats.total
        )
        options.worker.cache_memory_limit = self._calc_size_limit(
            self.args.cache_mem or options.worker.cache_memory_limit, mem_stats.total
        )
        options.worker.disk_limit = self.args.disk
        if self.args.spill_dir:
            from .spill import parse_spill_dirs
            options.worker.spill_directory = parse_spill_dirs(self.args.spill_dir)
            spill_dir_count = 1
        else:
            options.worker.spill_directory = None
            spill_dir_count = 0
        options.worker.advertise_addr = self.args.advertise

        self.n_process = 1 + options.worker.cpu_process_count + options.worker.io_process_count + spill_dir_count

        # start plasma
        self.start_plasma(options.worker.cache_memory_limit,
                          one_mapped_file=options.worker.plasma_one_mapped_file or False)

        kwargs['distributor'] = WorkerDistributor(self.n_process)
        return super(WorkerApplication, self).create_pool(*args, **kwargs)

    def start_service(self):
        super(WorkerApplication, self).start(self.endpoint, self.args.schedulers,
                                             self.pool, ignore_avail_mem=self.args.ignore_avail_mem)

    def handle_process_down(self, proc_indices):
        logger.debug('Process %r halt. Trying to recover.', proc_indices)
        for pid in proc_indices:
            self.pool.restart_process(pid)
        self._daemon_ref.handle_process_down(proc_indices, _tell=True)

    def stop_service(self):
        super(WorkerApplication, self).stop()


main = WorkerApplication()
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorNegativeBinomial(operands.NegativeBinomial, TensorRandomOperandMixin):
    __slots__ = '_n', '_p', '_size'
    _input_fields_ = ['_n', '_p']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorNegativeBinomial, self).__init__(_size=size, _state=state, _dtype=dtype,
                                                     _gpu=gpu, **kw)

    def __call__(self, n, p, chunk_size=None):
        return self.new_tensor([n, p], None, raw_chunk_size=chunk_size)


def negative_binomial(random_state, n, p, size=None, chunk_size=None, gpu=None, **kw):
    r"""
    Draw samples from a negative binomial distribution.

    Samples are drawn from a negative binomial distribution with specified
    parameters, `n` trials and `p` probability of success where `n` is an
    integer > 0 and `p` is in the interval [0, 1].

    Parameters
    ----------
    n : int or array_like of ints
        Parameter of the distribution, > 0. Floats are also accepted,
        but they will be truncated to integers.
    p : float or array_like of floats
        Parameter of the distribution, >= 0 and <=1.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``n`` and ``p`` are both scalars.
        Otherwise, ``np.broadcast(n, p).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized negative binomial distribution,
        where each sample is equal to N, the number of trials it took to
        achieve n - 1 successes, N - (n - 1) failures, and a success on the,
        (N + n)th trial.

    Notes
    -----
    The probability density for the negative binomial distribution is

    .. math:: P(N;n,p) = \binom{N+n-1}{n-1}p^{n}(1-p)^{N},

    where :math:`n-1` is the number of successes, :math:`p` is the
    probability of success, and :math:`N+n-1` is the number of trials.
    The negative binomial distribution gives the probability of n-1
    successes and N failures in N+n-1 trials, and success on the (N+n)th
    trial.

    If one throws a die repeatedly until the third time a "1" appears,
    then the probability distribution of the number of non-"1"s that
    appear before the third "1" is a negative binomial distribution.

    References
    ----------
    .. [1] Weisstein, Eric W. "Negative Binomial Distribution." From
           MathWorld--A Wolfram Web Resource.
           http://mathworld.wolfram.com/NegativeBinomialDistribution.html
    .. [2] Wikipedia, "Negative binomial distribution",
           http://en.wikipedia.org/wiki/Negative_binomial_distribution

    Examples
    --------
    Draw samples from the distribution:

    A real world example. A company drills wild-cat oil
    exploration wells, each with an estimated probability of
    success of 0.1.  What is the probability of having one success
    for each successive well, that is what is the probability of a
    single success after drilling 5 wells, after 6 wells, etc.?

    >>> import mars.tensor as mt

    >>> s = mt.random.negative_binomial(1, 0.1, 100000)
    >>> for i in range(1, 11):
    ...    probability = (mt.sum(s<i) / 100000.).execute()
    ...    print i, "wells drilled, probability of one success =", probability
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().negative_binomial(
            handle_array(n), handle_array(p), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorNegativeBinomial(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(n, p, chunk_size=chunk_size)
